from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cache

import mpmath
import numpy as np
from scipy.special import factorial

from osaft.core.backgroundfields import BackgroundField, WaveType
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import conj, full_range, integrate, pi
from osaft.core.variable import ActiveListVariable, ActiveVariable


class BaseGeneralARF(ABC):
    """Abstract base class for the computation of the acoustic radiation
    force for particle radius and  arbitrary boundary layer thickness
    """

    # Required attributes by child class (Active Variables)
    _x: ActiveVariable
    _x_v: ActiveVariable
    _alpha_n: ActiveVariable
    _beta_n: ActiveVariable
    field: BackgroundField

    # Required properties by child class
    abs_pos: float
    background_streaming: bool
    k_f: complex
    N_max: int
    rho_f: float
    wave_type: WaveType
    x: complex
    x_v: complex

    def __init__(self) -> None:
        """Constructor method"""

        self._D_n = ActiveListVariable(
            self._compute_D_n,
            "ARF coefficient D_n",
        )

        self._S_1n = ActiveListVariable(
            self._compute_S_1n,
            "ARF coefficient S_{1n}",
        )

        self._S_2n = ActiveListVariable(
            self._compute_S_2n,
            "ARF coefficient S_{2n}",
        )

        self._S_3n = ActiveListVariable(
            self._compute_S_3n,
            "ARF coefficient S_{3n}",
        )

        self._S_4n = ActiveListVariable(
            self._compute_S_4n,
            "ARF coefficient S_{4n}",
        )

        self._S_5n = ActiveListVariable(
            self._compute_S_5n,
            "ARF coefficient S_{5n}",
        )

        self._S_6n = ActiveListVariable(
            self._compute_S_6n,
            "ARF coefficient S_{6n}",
        )

        self._S_7n = ActiveListVariable(
            self._compute_S_7n,
            "ARF coefficient S_{7n}",
        )

        self._S_8n = ActiveListVariable(
            self._compute_S_8n,
            "ARF coefficient S_{8n}",
        )

        self._S_9n = ActiveListVariable(
            self._compute_S_9n,
            "ARF coefficient S_{9n}",
        )

        self._D_n.is_computed_by(
            self._alpha_n,
            self._beta_n,
            self._S_1n,
            self._S_2n,
            self._S_3n,
            self._S_4n,
            self._S_5n,
            self._S_6n,
            self._S_7n,
            self._S_8n,
            self._S_9n,
        )
        self._S_1n.is_computed_by(self._x, self._x_v)
        self._S_2n.is_computed_by(self._x, self._x_v)
        self._S_3n.is_computed_by(self._x, self._x_v)
        self._S_4n.is_computed_by(self._x, self._x_v)
        self._S_5n.is_computed_by(self._x, self._x_v)
        self._S_6n.is_computed_by(self._x, self._x_v)
        self._S_7n.is_computed_by(self._x, self._x_v)
        self._S_8n.is_computed_by(self._x, self._x_v)
        self._S_9n.is_computed_by(self._x, self._x_v)

    # -------------------------------------------------------------------------
    # Acoustic radiation force
    # -------------------------------------------------------------------------

    def _general_solution(self) -> float:
        if self.wave_type == WaveType.STANDING:
            return self._standing_general_solution()
        else:
            return self._travelling_general_solution()

    def _F_1(self):
        # F1 according to Eq. (5.15) for the rigid particle
        def term(n: int):
            term_1 = self.D_n(n) * self.field.A_in(n)
            term_1 *= conj(self.field.A_in(n + 1))
            term_2 = conj(self.D_n(n)) * self.field.A_in(n + 1)
            term_2 *= conj(self.field.A_in(n))
            prefactor = (n + 1) / (2 * n + 1) / (2 * n + 3)
            return prefactor * (term_1 + term_2)

        F1 = (
            -3
            / 2
            * pi
            * self.rho_f
            * sum(
                [term(n) for n in full_range(self.N_max)],
            )
        )
        return F1.real

    def _standing_general_solution(self) -> float:
        F1 = self._F_1()
        F2 = self._F_2_standing() if self.background_streaming else 0
        return F1 + F2

    def _travelling_general_solution(self) -> float:
        F1 = self._F_1()
        F2 = self._F_2_travelling() if self.background_streaming else 0
        return F1 + F2

    # -------------------------------------------------------------------------
    # Abstract Methods
    # -------------------------------------------------------------------------

    @abstractmethod
    def alpha_n(self, n: int) -> complex:
        pass

    @abstractmethod
    def beta_n(self, n: int) -> complex:
        pass

    @abstractmethod
    def _F_2_standing(self) -> float:
        pass

    @abstractmethod
    def _F_2_travelling(self) -> float:
        pass

    # -------------------------------------------------------------------------
    # D_n
    # -------------------------------------------------------------------------

    def D_n(self, n: int) -> complex:
        """coefficient :math:`D_{n}`

        :param n: order
        """
        return self._D_n.item(n)

    def _compute_D_n(self, n: int) -> None:
        # computation according to appendix (5.6)

        n1 = n + 1
        # first term
        out = self.S_1n(n) * self.alpha_n(n) * conj(self.alpha_n(n1))
        out += self.S_2n(n) * self.beta_n(n) * conj(self.beta_n(n1))
        out -= self.S_3n(n) * self.alpha_n(n) * conj(self.beta_n(n1))
        out -= self.S_4n(n) * self.beta_n(n) * conj(self.alpha_n(n1))
        out += self.S_5n(n) * self.alpha_n(n)
        out += self.S_6n(n) * conj(self.alpha_n(n1))
        out -= self.S_7n(n) * self.beta_n(n)
        out -= self.S_8n(n) * conj(self.beta_n(n1))
        out += self.S_9n(n) if self.background_streaming else 0

        return out

    # ------------------------------------------------------------------------
    # G_n_l
    # ------------------------------------------------------------------------

    @staticmethod
    def G_n_l(n: int, l: int, x: complex) -> complex:
        """Coefficient :math:`G_{n}^{(l)}(x)` from the appendix

        :param n: order
        :param l: kind of Hankel function
        """
        x_con = conj(x)
        if l == 1:
            tmp1 = Bf.d1_hankelh1(n, x_con)
            tmp2 = Bf.d1_hankelh1(n + 1, x_con)
        elif l == 2:
            tmp1 = Bf.d1_hankelh2(n, x_con)
            tmp2 = Bf.d1_hankelh2(n + 1, x_con)
        else:
            raise ValueError(f"l can only be 1 or 2 nor {l}")

        out = x_con * Bf.d1_hankelh1(n, x) * tmp1
        out += x * Bf.d1_hankelh1(n + 1, x) * tmp2

        return out

    def G_sum_1(self, n):
        """Numerically stable evaluation of
        :math:`(G_{n}^{(1)}(x) + G_{n}^{(2)}(x)) / 2`

        :param n: order
        """
        x = self.x
        x_con = self.x.conjugate()
        bessel_n = Bf.d1_besselj(n, x_con)
        bessel_n1 = Bf.d1_besselj(n + 1, x_con)
        term_1 = x_con * Bf.d1_hankelh1(n, x) * bessel_n
        term_2 = x * Bf.d1_hankelh1(n + 1, x) * bessel_n1
        return term_1 + term_2

    def G_sum_2(self, n):
        """Numerically stable evaluation of
        :math:`(G_{n}^{(1)*}(x^*) + G_{n}^{(2)}(x)) / 2`

        :param n: order
        """
        x = self.x
        x_con = self.x.conjugate()
        bessel_n = Bf.d1_besselj(n, x)
        bessel_n1 = Bf.d1_besselj(n + 1, x)
        term_1 = x_con * Bf.d1_hankelh1(n, x_con) * bessel_n
        term_2 = x * Bf.d1_hankelh1(n + 1, x_con) * bessel_n1
        return -(term_1 + term_2)

    # ------------------------------------------------------------------------
    # K_n_l
    # ------------------------------------------------------------------------

    def K_n_l(self, n: int, l: int) -> complex:
        """Coefficient :math:`K_{n}^{(l)}(x)` from the appendix

        :param n: order
        :param l: kind of Hankel function
        """
        x_con = self.x.conjugate()
        if l == 1:
            tmp1 = Bf.d1_hankelh1(n, x_con)
            tmp2 = Bf.d1_hankelh1(n + 1, x_con)
        elif l == 2:
            tmp1 = Bf.d1_hankelh2(n, x_con)
            tmp2 = Bf.d1_hankelh2(n + 1, x_con)
        else:
            raise ValueError(f"l can only be 1 or 2 nor {l}")

        out = (n + 1) * x_con**2 / self.x_v**2 * tmp1
        out *= Bf.hankelh1(n, self.x_v)

        out -= x_con * Bf.d1_hankelh1(n + 1, self.x_v) * tmp2
        return out

    def K_sum(self, n: int):
        """Numerically stable evaluation of
        :math:`(K_{n}^{(1)}(x) + K_{n}^{(2)}(x)) / 2`

        :param n: order
        """
        x_v = self.x_v
        x_con = self.x.conjugate()
        prefactor = (n + 1) * x_con**2 / x_v**2
        term_1 = Bf.hankelh1(n, x_v) * Bf.d1_besselj(n, x_con)
        term_2 = x_con * Bf.d1_hankelh1(n + 1, x_v)
        term_2 *= Bf.d1_besselj(n + 1, x_con)
        return prefactor * term_1 - term_2

    # ------------------------------------------------------------------------
    # L_n_l
    # ------------------------------------------------------------------------

    def L_n_l(self, n: int, l: int) -> complex:
        """Numerically stable evaluation of
        :math:`(K_{n}^{(1)}(x) + K_{n}^{(2)}(x)) / 2`

        :param n: order
        :param l: kind of Hankel function
        """
        x_v_con = self.x_v.conjugate()
        if l == 1:
            tmp1 = Bf.d1_hankelh1(n, self.x)
            tmp2 = Bf.d1_hankelh1(n + 1, self.x)
        elif l == 2:
            tmp1 = Bf.d1_hankelh2(n, self.x)
            tmp2 = Bf.d1_hankelh2(n + 1, self.x)
        else:
            raise ValueError(f"l can only be 1 or 2 nor {l}")

        out = -(n + 1) * self.x**2 / self.x_v**2 * tmp2
        out *= Bf.hankelh2(n + 1, x_v_con)

        out += self.x * Bf.d1_hankelh2(n, x_v_con) * tmp1
        return out

    def L_sum(self, n: int) -> complex:
        """Numerically stable evaluation of
        :math:`(L_{n}^{(1)}(x) + L_{n}^{(2)}(x)) / 2`

        :param n: order
        """
        x = self.x
        x_v = self.x_v
        x_v_con = self.x_v.conjugate()
        term_1 = x * Bf.d1_hankelh2(n, x_v_con) * Bf.d1_besselj(n, x)
        term_2 = (n + 1) * x**2 / x_v**2 * Bf.hankelh2(n + 1, x_v_con)
        term_2 *= Bf.d1_besselj(n + 1, x)
        return term_1 - term_2

    # -------------------------------------------------------------------------
    # H_nm, J_nm
    # -------------------------------------------------------------------------

    @staticmethod
    def B_kq(k: int, q: int, m: int, n: int) -> complex:
        """Coefficient :math:`B_{kq}` from the appendix

        :param k: :math:`k`
        :param q: :math:`q`
        :param m: :math:`m`
        :param n: :math:`n`
        """
        numerator = 1j ** (m - n) * (-1) ** k
        numerator *= factorial(n + k) * factorial(m + q)
        denominator = (2 * 1j) ** (k + q) * factorial(k) * factorial(n - k)
        denominator *= factorial(q) * factorial(m - q)
        return numerator / denominator

    @cache
    def J_nm_j(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ) -> complex:
        """Integral :math:`J_{nm}^{(j)}`

        :param x1: :math:`x_1`
        :param x2: :math:`x_2`
        :param n: :math:`n`
        :param m: :math:`m`
        :param j: :math:`j`
        """
        if 1.01 * abs(x1.real) > abs(x1.imag) > 0.99 * abs(x1.real):
            return self._J_nm_j_quadrature(x1, x2, n, m, j)
        else:
            return self._J_nm_j_sum(x1, x2, n, m, j)

    @cache
    def H_nm_j(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ) -> complex:
        """Integral :math:`H_{nm}^{(j)}`

        :param x1: :math:`x_1`
        :param x2: :math:`x_2`
        :param n: :math:`n`
        :param m: :math:`m`
        :param j: :math:`j`
        """
        return self.H_nm_j_sum(x1, x2, n, m, j)

    @cache
    def H_nm_j_sum(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ) -> complex:
        """Coefficient :math:`H_{nm}^{(j)}`

        :param x1: argument x1
        :param x2: argument x2
        :param n: order n
        :param m: order m
        :param j: exponent
        """
        out = mpmath.mpc(0)
        arg_1 = mpmath.mpc(1j * x2.conjugate() - 1j * x1)
        for k in full_range(n):
            for q in full_range(m):
                M = 2 + k + q + j
                term = mpmath.expint(M, arg_1)
                term *= mpmath.mpc(x1 ** (-(k + 1)))
                term *= mpmath.mpc(x2.conjugate() ** (-(q + 1)))
                term *= self.B_kq(k, q, m, n)
                out += term
        return complex(out)

    def _J_nm_j_sum(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ) -> complex:
        """Integrate J_nm using sum of Bessel functions

        :param x1: :math:`x_1`
        :param x2: :math:`x_2`
        :param n: :math:`n`
        :param m: :math:`m`
        :param j: :math:`j`
        """
        arg_1 = mpmath.mpc(-1j * x1 + 1j * x2.conjugate())
        arg_2 = mpmath.mpc(-1j * x1 - 1j * x2.conjugate())
        out = mpmath.mpc(0)
        for k in full_range(n):
            for q in full_range(m):
                M = 2 + k + q + j
                term = mpmath.expint(M, arg_1)
                term -= (-1) ** (q + m) * mpmath.expint(M, arg_2)
                term *= mpmath.mpc(x1 ** (-(k + 1)))
                term *= mpmath.mpc(x2.conjugate() ** (-(q + 1)))
                term *= self.B_kq(k, q, m, n)
                out += term
        return complex(0.5 * out)

    def _J_nm_j_quadrature(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ) -> complex:
        """Integrate J_nm using quadrature

        :param x1: :math:`x_1`
        :param x2: :math:`x_2`
        :param n: :math:`n`
        :param m: :math:`m`
        :param j: :math:`j`
        """

        def kernel(y: float) -> complex:
            out = y ** (-j) * Bf.hankelh1(n, x1 * y)
            out *= Bf.besselj(m, x2.conjugate() * y)
            return out

        return self._integrate_J_nm_H_nm(kernel)

    @staticmethod
    def _integrate_J_nm_H_nm(kernel):
        """Integrate J_nm or H_nm using quadrature

        :param kernel: kernel for integration
        """

        def real_kernel(y: float) -> float:
            return kernel(y).real

        def imag_kernel(y: float) -> float:
            return kernel(y).imag

        real_part, _ = integrate(real_kernel, 1, np.inf)
        imag_part, _ = integrate(imag_kernel, 1, np.inf)

        return real_part + 1j * imag_part

    # ------------------------------------------------------------------------
    # S_in
    # ------------------------------------------------------------------------

    def S_1n(self, n: int) -> complex:
        """coefficient :math:`S_{1n}`

        :param n: order
        """
        return self._S_1n.item(n)

    def _compute_S_1n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = x.conjugate()
        x_v = self.x_v
        n1 = n + 1
        # first term
        term_1 = self.H_nm_j(x, x, n, n1, 0)
        term_1 -= self.H_nm_j(x, x, n, n1, 2)
        term_1 *= ((n + 2) * x**2 - n * x_con**2) / 2

        # second term
        term_2 = self.H_nm_j(x, x, n1, n, 0)
        term_2 -= self.H_nm_j(x, x, n1, n, 2)
        term_2 *= x * x_con

        # third term
        term_3 = x * x_con / x_v**2 * self.G_n_l(n, 2, x)

        # last term
        term_4 = n * x_con * self.H_nm_j(x, x, n, n, 1)
        term_4 += (n + 2) * x * self.H_nm_j(x, x, n1, n1, 1)
        term_4 -= x_con * x * self.H_nm_j(x, x, n1, n, 0)
        term_4 *= (x**2 - x_con**2) / x_v**2

        return term_1 - term_2 + term_3 + term_4

    def S_2n(self, n: int) -> complex:
        """coefficient :math:`S_{2n}`

        :param n: order
        """
        return self._S_2n.item(n)

    def _compute_S_2n(self, n: int) -> complex:
        # computation according to appendix A
        x_v = self.x_v
        x_v_con = x_v.conjugate()
        n1 = n + 1
        # first term
        term_1 = self.H_nm_j(x_v, x_v, n1, n, 0)
        term_1 -= self.H_nm_j(x_v, x_v, n1, n, 2)
        term_1 *= x_v * x_v_con
        # second term
        term_2 = self.H_nm_j(x_v, x_v, n, n1, 0)
        term_2 -= self.H_nm_j(x_v, x_v, n, n1, 2)
        term_2 *= n1 * x_v**2

        # third term
        term_3 = Bf.d1_hankelh1(n1, x_v) * Bf.hankelh2(n1, x_v_con)
        term_3 += Bf.hankelh1(n, x_v) * Bf.d1_hankelh2(n, x_v_con)
        term_3 *= n1

        return n * (n + 2) * (term_1 - term_2 - term_3)

    def S_3n(self, n: int) -> complex:
        """coefficient :math:`S_{3n}`

        :param n: order
        """
        return self._S_3n.item(n)

    def _compute_S_3n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_v = self.x_v
        x_v_con = x_v.conjugate()
        n1 = n + 1

        # first term
        term_1 = self.H_nm_j(x, x_v, n, n1, 0)
        term_1 -= self.H_nm_j(x, x_v, n, n1, 2)
        term_1 *= (n * x_v_con**2 - n1 * x**2) / 2

        # second term
        term_2 = self.H_nm_j(x, x_v, n1, n, 0)
        term_2 -= self.H_nm_j(x, x_v, n1, n, 2)
        term_2 *= x * x_v_con

        # third term
        term_3 = self.H_nm_j(x, x_v, n, n, -1)
        term_3 -= self.H_nm_j(x, x_v, n, n, 1)
        term_3 *= x**2 * x_v_con / 2

        # fourth term
        term_4 = self.H_nm_j(x, x_v, n1, n1, -1)
        term_4 -= self.H_nm_j(x, x_v, n1, n1, 1)
        term_4 *= x * x_v**2 / 2

        # fifth term
        term_5 = n * x_v_con * self.H_nm_j(x, x_v, n, n, 1)
        term_5 -= n1 * x * self.H_nm_j(x, x_v, n1, n1, 1)
        term_5 *= x**2 / x_v**2

        # sixth term
        term_6 = self.L_n_l(n, 1)

        return (n + 2) * (term_1 - term_2 + term_3 + term_4 + term_5 - term_6)

    def S_4n(self, n: int) -> complex:
        """coefficient :math:`S_{4n}`

        :param n: order
        """
        return self._S_4n.item(n)

    def _compute_S_4n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = x.conjugate()
        x_v = self.x_v
        n1 = n + 1
        # first term
        term_1 = self.H_nm_j(x_v, x, n, n1, 0)
        term_1 -= self.H_nm_j(x_v, x, n, n1, 2)
        term_1 *= ((n + 2) * x_v**2 - n1 * x_con**2) / 2

        # second term
        term_2 = self.H_nm_j(x_v, x, n1, n, 0)
        term_2 -= self.H_nm_j(x_v, x, n1, n, 2)
        term_2 *= x_con * x_v

        # third term
        term_3 = self.H_nm_j(x_v, x, n, n, -1)
        term_3 -= self.H_nm_j(x_v, x, n, n, 1)
        term_3 *= x_con * x_v**2 / 2

        # fourth term
        term_4 = self.H_nm_j(x_v, x, n1, n1, -1)
        term_4 -= self.H_nm_j(x_v, x, n1, n1, 1)
        term_4 *= x_con**2 * x_v / 2

        # fifth term
        term_5 = (n + 2) * x_v * self.H_nm_j(x_v, x, n1, n1, 1)
        term_5 = n1 * x_con * self.H_nm_j(x_v, x, n, n, 1) - term_5
        term_5 *= x_con**2 / x_v**2

        # sixth term
        term_6 = self.K_n_l(n, 2)

        return n * (term_1 + term_2 - term_3 + term_4 - term_5 + term_6)

    def S_5n(self, n: int) -> complex:
        """coefficient :math:`S_{5n}`

        :param n: order
        """
        return self._S_5n.item(n)

    def _compute_S_5n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = x.conjugate()
        x_v = self.x_v
        n1 = n + 1

        # first term
        term_1 = self.J_nm_j(x, x, n, n1, 0)
        term_1 -= self.J_nm_j(x, x, n, n1, 2)
        term_1 *= 0.5 * ((n + 2) * x**2 - n * x_con**2)

        # second term
        term_2 = self.J_nm_j(x, x, n1, n, 0)
        term_2 -= self.J_nm_j(x, x, n1, n, 2)
        term_2 *= x_con * x

        # third term
        term_3 = self.G_sum_1(n)
        term_3 *= x_con * x / x_v**2

        # fourth term
        term_4 = n * x_con * self.J_nm_j(x, x, n, n, 1)
        term_4 += (n + 2) * x * self.J_nm_j(x, x, n1, n1, 1)
        term_4 -= x_con * x * self.J_nm_j(x, x, n1, n, 0)
        term_4 *= (x**2 - x_con**2) / x_v**2

        return term_1 - term_2 + term_3 + term_4

    def S_6n(self, n: int) -> complex:
        """coefficient :math:`S_{6n}`

        :param n: order
        """
        self._compute_S_6n(n)
        return self._S_6n.item(n)

    def _compute_S_6n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = x.conjugate()
        x_v = self.x_v
        n1 = n + 1
        # first term
        term_1 = self.J_nm_j(x, x, n1, n, 0)
        term_1 -= self.J_nm_j(x, x, n1, n, 2)
        term_1 = 0.5 * ((n + 2) * x**2 - n * x_con**2) * conj(term_1)

        # second term
        term_2 = self.J_nm_j(x, x, n, n1, 0)
        term_2 -= self.J_nm_j(x, x, n, n1, 2)
        term_2 = x_con * x * conj(term_2)

        # third term
        term_3 = self.G_sum_2(n)
        term_3 *= x_con * x / x_v**2

        # fourth term
        term_4 = n * x * self.J_nm_j(x, x, n, n, 1)
        term_4 += (n + 2) * x_con * self.J_nm_j(x, x, n1, n1, 1)
        term_4 -= x_con * x * self.J_nm_j(x, x, n, n1, 0)
        term_4 = (x**2 - x_con**2) / x_v**2 * conj(term_4)

        return term_1 - term_2 + term_3 + term_4

    def S_7n(self, n: int) -> complex:
        """coefficient :math:`S_{7n}`

        :param n: order
        """
        return self._S_7n.item(n)

    def _compute_S_7n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = x.conjugate()
        x_v = self.x_v
        n1 = n + 1

        # first term
        term_1 = self.J_nm_j(x_v, x, n, n1, 0)
        term_1 -= self.J_nm_j(x_v, x, n, n1, 2)
        term_1 *= (n + 2) * x_v**2 - n1 * x_con**2
        term_1 *= 0.5

        # second term
        term_2 = self.J_nm_j(x_v, x, n1, n, 0)
        term_2 -= self.J_nm_j(x_v, x, n1, n, 2)
        term_2 *= x_con * x_v

        # third term
        term_3 = self.J_nm_j(x_v, x, n, n, -1)
        term_3 -= self.J_nm_j(x_v, x, n, n, 1)
        term_3 *= 0.5 * x_con * x_v**2

        # fourth term
        term_4 = self.J_nm_j(x_v, x, n1, n1, -1)
        term_4 -= self.J_nm_j(x_v, x, n1, n1, 1)
        term_4 *= 0.5 * x_con**2 * x_v

        # fifth term
        term_5 = self.K_sum(n)

        # sixth term
        term_6 = -(n + 2) * x_v * self.J_nm_j(x_v, x, n1, n1, 1)
        term_6 += n1 * x_con * self.J_nm_j(x_v, x, n, n, 1)
        term_6 *= x_con**2 / x_v**2

        return n * (term_1 + term_2 - term_3 + term_4 + term_5 - term_6)

    def S_8n(self, n: int) -> complex:
        """coefficient :math:`S_{8n}`

        :param n: order
        """
        return self._S_8n.item(n)

    def _compute_S_8n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = self.x.conjugate()
        x_v = self.x_v
        x_v_con = x_v.conjugate()
        n1 = n + 1
        # first term
        term_1 = self.J_nm_j(x_v, x, n1, n, 0)
        term_1 -= self.J_nm_j(x_v, x, n1, n, 2)
        term_1 = conj(term_1) * (x_v_con**2 * n - n1 * x**2)
        term_1 *= 0.5

        # second term
        term_2 = self.J_nm_j(x_v, x, n, n1, 0)
        term_2 -= self.J_nm_j(x_v, x, n, n1, 2)
        term_2 = x_v_con * x * conj(term_2)

        # third term
        term_3 = self.J_nm_j(x_v, x, n, n, -1)
        term_3 -= self.J_nm_j(x_v, x, n, n, 1)
        term_3 = 0.5 * x_v_con * x**2 * conj(term_3)

        # fourth term
        term_4 = self.J_nm_j(x_v, x, n1, n1, -1)
        term_4 -= self.J_nm_j(x_v, x, n1, n1, 1)
        term_4 = 0.5 * x_v**2 * x * conj(term_4)

        # fifth term
        term_5 = self.L_sum(n)

        # sixth term
        term_6 = n * x_v * self.J_nm_j(x_v, x, n, n, 1)
        term_6 -= n1 * x_con * self.J_nm_j(x_v, x, n1, n1, 1)
        term_6 = x**2 / x_v**2 * conj(term_6)

        # common factor
        return (n + 2) * (term_1 - term_2 + term_3 + term_4 - term_5 + term_6)

    def S_9n(self, n: int) -> complex:
        """coefficient :math:`S_{9n}`

        :param n: order
        """
        return self._S_9n.item(n)

    def _compute_S_9n(self, n: int) -> complex:
        # computation according to appendix A
        x = self.x
        x_con = x.conjugate()
        x_v = self.x_v
        n1 = n + 1

        out = x * Bf.d1_besselj(n1, x) * Bf.d1_besselj(n1, x_con)
        out += x_con * Bf.d1_besselj(n, x) * Bf.d1_besselj(n, x_con)

        return out * x * x_con / x_v**2


if __name__ == "__main__":
    pass



from __future__ import annotations

from abc import ABC

import numpy as np

from osaft import ViscoelasticFluid, ViscousFluid, WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveListVariable
from osaft.solutions.basedoinikov2021.base import BaseDoinikov

NDArray = np.ndarray


class CoefficientMatrix(BaseDoinikov, ABC):
    """Coefficient matrix class for scattering field class for Doinikov
    (viscous fluid-elastic sphere; 2021)

    :param f: frequency [Hz]
    :param R_0: radius [m]
    :param rho_s: density of the particle [kg/m^3]
    :param E_s: Young's modulus of the particle [Pa]
    :param nu_s: Poisson's ratio of the particle [-]
    :param fluid: Fluid instance of the model
    :param p_0: pressure amplitude [Pa]
    :param wave_type: wave type
    :param position: position in the standing wave field [m]
    """

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float,
        rho_s: float,
        E_s: float,
        nu_s: float,
        fluid: ViscousFluid | ViscoelasticFluid,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
    ) -> None:
        """Constructor method"""

        super().__init__(
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            E_s=E_s,
            nu_s=nu_s,
            fluid=fluid,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
        )
        self._list_matrix_M_n = ActiveListVariable(
            self._compute_matrix_M,
            "Matrices M(n)",
        )

        self._list_matrix_M_n.is_computed_by(
            self._x_f,
            self._x_v,
            self._x_l,
            self._x_t,
        )

        self._list_vector_n = ActiveListVariable(
            self._compute_vector_n,
            "Vectors N(n)",
        )
        self._list_vector_n.is_computed_by(
            self._x_f,
            self.field._A_in,
        )

    def det_M_n(self, n: int, column: None | int = None) -> complex:
        """Determinant of the matrix `M` for the mode `n`

        :param n: mode
        :param column: the `l`th coefficient is replaced with the vector `N`
        """
        matrix = self.matrix_M(n).copy()
        vector = self.vector_n(n)
        if column is not None:
            matrix[:, column] = vector
        return np.linalg.det(matrix)

    def matrix_M(self, n: int) -> NDArray:
        """Matrix :math:`M_l`

        (Eq. A19 - A 22)

        :param n: mode
        """
        return self._list_matrix_M_n.item(n)

    def vector_n(self, n: int) -> NDArray:
        """Vector :math:`n_l`

        (Eq. A19 - A 22)

        :param n: mode
        """
        return self._list_vector_n.item(n)

    def _compute_matrix_M(self, l: int) -> NDArray:
        """Matrix :math:`M_l`

        (Eq. A19 - A 22)

        :param l: mode
        :return: matrix M_l
        """
        M = np.zeros((4, 4), dtype=complex)

        # First Row
        M[0, 0] = self.x_f * Bf.d1_hankelh1(l, self.x_f)
        M[0, 1] = -l * (l + 1) * Bf.hankelh1(l, self.x_v)
        M[0, 2] = 1j * self.omega * self.x_l * Bf.d1_besselj(l, self.x_l)
        M[0, 3] = -1j * self.omega * l * (l + 1) * Bf.besselj(l, self.x_t)
        M[1, 0] = Bf.hankelh1(l, self.x_f)
        M[1, 1] = -Bf.hankelh1(l, self.x_v)
        M[1, 1] -= self.x_v * Bf.d1_hankelh1(l, self.x_v)
        M[1, 2] = 1j * self.omega * Bf.besselj(l, self.x_l)
        M[1, 3] = Bf.besselj(l, self.x_t)
        M[1, 3] += self.x_t * Bf.d1_besselj(l, self.x_t)
        M[1, 3] *= -1j * self.omega

        # Second Column
        M[2, 0] = 2 * self.eta * Bf.d2_hankelh1(l, self.x_f)
        M[2, 0] -= self.viscosity_term() * Bf.hankelh1(l, self.x_f)
        M[2, 0] *= self.x_f**2
        M[2, 1] = Bf.hankelh1(l, self.x_v)
        M[2, 1] -= self.x_v * Bf.d1_hankelh1(l, self.x_v)
        M[2, 1] *= 2 * l * (l + 1) * self.eta
        M[2, 2] = -Bf.d2_besselj(l, self.x_l)
        M[2, 2] += self.nu_s * Bf.besselj(l, self.x_l) / (1 - 2 * self.nu_s)
        M[2, 2] *= self.E_s * self.x_l**2 / (1 + self.nu_s)
        M[2, 3] = self.x_t * Bf.d1_besselj(l, self.x_t)
        M[2, 3] -= Bf.besselj(l, self.x_t)
        M[2, 3] *= l * (l + 1) * self.E_s / (1 + self.nu_s)

        # Third Column
        M[3, 0] = self.x_f * Bf.d1_hankelh1(l, self.x_f)
        M[3, 0] -= Bf.hankelh1(l, self.x_f)
        M[3, 0] *= 2 * self.eta
        M[3, 1] = (2 - l**2 - l) * Bf.hankelh1(l, self.x_v)
        M[3, 1] -= self.x_v**2 * Bf.d2_hankelh1(l, self.x_v)
        M[3, 1] *= self.eta
        M[3, 2] = Bf.besselj(l, self.x_l)
        M[3, 2] -= self.x_l * Bf.d1_besselj(l, self.x_l)
        M[3, 2] *= self.E_s / (1 + self.nu_s)
        M[3, 3] = self.x_t**2 * Bf.d2_besselj(l, self.x_t)
        M[3, 3] -= (2 - l**2 - l) * Bf.besselj(l, self.x_t)
        M[3, 3] *= self.E_s / (2 * (1 + self.nu_s))
        return M

    def _compute_vector_n(self, l: int) -> NDArray:
        """Vector :math:`n_l`

        (Eq. A19 - A 22)

        :param l: mode
        :return: matrix M_l
        """
        vec_n = np.zeros(4, dtype=complex)

        vec_n[0] = -self.A_in(l) * self.x_f * Bf.d1_besselj(l, self.x_f)
        vec_n[1] = -self.A_in(l) * Bf.besselj(l, self.x_f)
        vec_n[2] = self.viscosity_term() * Bf.besselj(l, self.x_f)
        vec_n[2] -= 2 * self.eta * Bf.d2_besselj(l, self.x_f)
        vec_n[2] *= self.A_in(l) * self.x_f**2
        vec_n[3] = Bf.besselj(l, self.x_f)
        vec_n[3] -= self.x_f * Bf.d1_besselj(l, self.x_f)
        vec_n[3] *= 2 * self.A_in(l) * self.eta

        return vec_n

    def viscosity_term(self) -> complex:
        """Often used function of fluids shear and bulk viscosity"""
        return (
            1j * self.rho_f * self.c_f**2 / self.omega
            + self.zeta
            - 2 * self.eta / 3
        )

    def a_0(self):
        """Coefficient :math:`a_0` [m^2/s]

        (Eq. A15)

        :return: coefficient a_0
        """
        term1 = Bf.d2_besselj(0, self.x_l)
        term1 -= self.nu_s * Bf.besselj(0, self.x_l) / (1 - 2 * self.nu_s)
        term1 *= self.E_s * self.k_f * self.x_l**2
        term1 *= Bf.besselj(1, self.x_f) / (1 + self.nu_s)

        term2 = 2 * self.eta * Bf.d2_besselj(0, self.x_f)
        term2 -= Bf.besselj(0, self.x_f) * self.viscosity_term()
        term2 *= 1j * self.omega * self.k_l * self.x_f**2
        term2 *= Bf.besselj(1, self.x_l)

        return (term1 + term2) * self.A_in(0) / self.D_0()

    def a_hat_0(self):
        """Coefficient :math:`\\hat{a}_0` [m^2/s]

        (Eq. A15)

        :return: coefficient a_0
        """
        prefactor = self.A_in(0) * self.k_f * self.x_f**2 / self.D_0()

        term1 = Bf.besselj(0, self.x_f) * self.viscosity_term()
        term1 -= 2 * self.eta * Bf.d2_besselj(0, self.x_f)
        term1 *= Bf.hankelh1(1, self.x_f)

        term2 = 2 * self.eta * Bf.d2_hankelh1(0, self.x_f)
        term2 -= Bf.hankelh1(0, self.x_f) * self.viscosity_term()
        term2 *= Bf.besselj(1, self.x_f)

        return prefactor * (term1 + term2)

    def D_0(self):
        """Coefficient :math:`D_0`

        (Eq. A17)

        :return: coefficient D_0
        """
        prefactor1 = self.E_s * self.k_f * self.x_l**2
        prefactor1 *= Bf.hankelh1(1, self.x_f) / (1 + self.nu_s)

        term1 = self.nu_s * Bf.besselj(0, self.x_l) / (1 - 2 * self.nu_s)
        term1 -= Bf.d2_besselj(0, self.x_l)

        prefactor2 = 1j * self.omega * self.k_l * self.x_f**2
        prefactor2 *= Bf.besselj(1, self.x_l)

        term2 = Bf.hankelh1(0, self.x_f) * self.viscosity_term()
        term2 -= 2 * self.eta * Bf.d2_hankelh1(0, self.x_f)

        return prefactor1 * term1 + prefactor2 * term2

    if __name__ == "__main__":
        pass



from osaft.solutions.doinikov2021viscous.arf import ARF
from osaft.solutions.doinikov2021viscous.scattering import ScatteringField
from osaft.solutions.doinikov2021viscous.streaming import StreamingField



import unittest

from osaft import WaveType, doinikov1994rigid, settnes2012
from osaft.tests.basetest_arf import HelperCompareARF
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestCompareSmallBoundaryLayer(BaseTestSolutions, HelperCompareARF):
    def setUp(self):
        super().setUp()

        self.arf_compare_threshold = 10e-2
        self.small_boundary_layer = True
        self.large_boundary_layer = False
        self.long_wavelength = True

        # Viscosity
        self.parameters._eta_f.high = 1e-3
        self.parameters._eta_f.low = 1e-5
        self.parameters.zeta_f = 0

        # Radius
        self.parameters._R_0.low = 10e-6
        self.parameters._R_0.high = 50e-6

        # Frequency
        self.parameters._f.low = 5e5
        self.parameters._f.high = 1e6

        # Density
        self.parameters._rho_f.low = 1500

        self.cls = SolutionFactory().doinikov_1994_rigid_arf()
        self.cls.long_wavelength = self.long_wavelength
        self.cls.large_boundary_layer = self.large_boundary_layer
        self.cls.small_boundary_layer = self.small_boundary_layer

        self.compare_cls = SolutionFactory().doinikov_1994_rigid_arf()
        self.compare_cls.long_wavelength = self.long_wavelength
        self.list_cls = [self.cls, self.compare_cls]
        self.assign_parameters()


class TestCompareSmallParticleGeneral(BaseTestSolutions):
    def setUp(self):
        super().setUp()

        self.arf_compare_threshold = 1e-1

        self.parameters._R_0.low = 1e-6
        self.parameters._R_0.high = 5e-6
        self.parameters._eta_f.low = 1e-6
        self.parameters._eta_f.high = 1e-3
        self.parameters._f.high = 5e5

        self.large_boundary_layer = False
        self.small_boundary_layer = False

        self.cls = doinikov1994rigid.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.p_0,
            WaveType.STANDING,
            self.position,
            long_wavelength=True,
        )

        self.compare_cls = doinikov1994rigid.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.p_0,
            WaveType.STANDING,
            self.position,
            long_wavelength=False,
        )

        self.list_cls = [self.cls, self.compare_cls]

    def test_with_background_streaming(self):
        self.cls.background_streaming = True
        self.compare_cls.background_streaming = True
        self.do_testing(
            self.cls.compute_arf,
            self.compare_cls.compute_arf,
            threshold=self.arf_compare_threshold,
            zero=1e-30,
        )

    def test_without_background_streaming(self):
        self.cls.background_streaming = False
        self.compare_cls.background_streaming = False
        self.parameters._wave_type.list_of_values = [WaveType.STANDING]
        self.do_testing(
            self.cls.compute_arf,
            self.compare_cls.compute_arf,
            threshold=self.arf_compare_threshold,
            zero=1e-30,
        )


class TestCompareSettnes(BaseTestSolutions, HelperCompareARF):
    def setUp(self):
        super().setUp()

        self.arf_compare_threshold = 10e-2
        self.small_boundary_layer = False
        self.large_boundary_layer = False
        self.long_wavelength = True

        # Viscosity
        self.parameters._eta_f.high = 1e-3
        self.parameters._eta_f.low = 1e-5
        self.parameters._eta_f.value = 1e-4
        self.parameters.zeta_f = 0

        # Radius
        self.parameters._R_0.low = 1e-6
        self.parameters._R_0.high = 10e-6

        # Frequency
        self.parameters._f.low = 1e6
        self.parameters._f.high = 2e6

        # Density
        self.parameters._rho_s.low = 1500
        self.parameters._rho_s.high = 2000

        # Speed of Sound Scatterer
        self.parameters._c_s.low = 1e6
        self.parameters._c_s.high = 1e6

        # Wave Type
        self.parameters._wave_type.list_of_values = [WaveType.STANDING]
        self.parameters.wave_type = WaveType.STANDING

        self.cls = doinikov1994rigid.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.zeta_f,
            self.p_0,
            self.wave_type,
            self.position,
            long_wavelength=self.long_wavelength,
            large_boundary_layer=self.large_boundary_layer,
            small_boundary_layer=self.small_boundary_layer,
            background_streaming=False,
        )

        self.compare_cls = settnes2012.ARF(
            self.f,
            self.R_0,
            self.rho_s,
            self.c_s,
            self.rho_f,
            self.c_f,
            self.eta_f,
            self.p_0,
            self.wave_type,
            self.position,
        )

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from collections.abc import Callable

import numpy as np
from matplotlib import colors
from matplotlib import pyplot as plt
from matplotlib import tri
from matplotlib.animation import FuncAnimation
from matplotlib.cm import ScalarMappable
from matplotlib.collections import TriMesh
from matplotlib.colorbar import Colorbar
from matplotlib.text import Annotation
from matplotlib.tri import Triangulation
from mpl_toolkits.axes_grid1 import make_axes_locatable

from osaft.core.functions import pi, sqrt
from osaft.plotting.scattering.base_plotter import BaseScatteringPlotter

NDArray = np.ndarray


class TriangulationPlotter(BaseScatteringPlotter):
    """Base class for plotting first order solutions such as the acoustic
    scattering fields and particle vibrations.

    :param is_scatterer: if `True`, plot of the scatterer,
        else plot of the fluid
    :param cmap: colormap passed to matplotlib
    :param div_cmap: diverging colormap passed to matplotlib
    """

    def __init__(
        self,
        is_scatterer: bool,
        cmap: str = "winter",
        div_cmap: str = "Spectral_r",
    ) -> None:
        """Constructor method"""
        # Call to parent class
        super().__init__(is_scatterer)

        # Color map
        self.cmap = cmap
        self.div_cmap = div_cmap

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def plot(
        self,
        X: NDArray,
        Y: NDArray,
        C: NDArray,
        radius: None | float,
        symmetric: bool,
        tripcolor: bool,
        cbar_label: str,
        use_diverging_cmap: bool,
        ax: None | plt.Axes,
        **kwargs,
    ) -> [
        None | plt.Figure,
        plt.Axes,
        TriMesh,
        None | Colorbar,
        tri.Triangulation,
    ]:
        """Plot for the scattering field

        Plots the scattering field using `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        or `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontourf.html>`_
        from Matplotlib.

        :param X: array of x-coordinates
        :param Y: array of y-coordinates
        :param C: array of the values
        :param radius: radius for masked triangulation
        :param symmetric: if true the plot is symmetric w.r.t. the x-axis
        :param tripcolor: if ``True`` tripcolor plot
        :param cbar_label: label for the color bar
        :param use_diverging_cmap: use the specified diverging cmap
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        """
        # Create figure and axis
        fig, ax = self._create_figure(ax)
        # Triangulation
        triang = self.masked_triangulation(X, Y, symmetric, radius)
        # Symmetry
        if symmetric:
            C = np.concatenate([C, C])

        if "cmap" not in kwargs.keys():
            cmap = self.div_cmap if use_diverging_cmap else self.cmap
            kwargs["cmap"] = cmap

        # Plot tripcolor or tricontourf plot
        if tripcolor:
            if "shading" not in kwargs.keys():
                kwargs["shading"] = "gouraud"

            trc = ax.tripcolor(triang, C, **kwargs)
        else:
            trc = ax.tricontourf(triang, C, **kwargs)

        # Add color bar, axis limits, axis labels, etc.
        cbar = self._finish_triplot(fig, ax, radius, trc, cbar_label)

        return fig, ax, trc, cbar, triang

    def animate(
        self,
        frames: int,
        interval: float,
        tripcolor: bool,
        symmetric: bool,
        cbar_label: str,
        use_diverging_cmap: bool,
        radius: float,
        animate_meth: Callable[[float], tuple[NDArray, NDArray, NDArray]],
        C_norm: NDArray,
        ax: None | plt.Axes,
        **kwargs,
    ) -> FuncAnimation:
        """Animates tricontourf and tripcolor plot

        :param frames: number of frames
        :param interval: interval between frames in ms
        :param tripcolor: if ``True`` tripcolor, else tricontourf
        :param symmetric: if the symmetric part of plot shall be plotted
        :param cbar_label: label for the color bar
        :param use_diverging_cmap: use as :attr:`div_cmap`
        :param animate_meth: function that returns speed for given phase
        :param C_norm: absolute values of the field to be plotted
        :param radius: radius
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tripcolor or tricontourf
        """
        # Get Coordinates
        X, Z, _ = animate_meth(0)

        vmax = 1.01 * np.abs(C_norm).max()
        if use_diverging_cmap:
            vmin = -vmax
        else:
            vmin = 0

        # Initial plot
        fig, ax, triplot, cbar, triang = self.plot(
            Z,
            X,
            C_norm,
            radius,
            symmetric,
            tripcolor,
            cbar_label,
            use_diverging_cmap,
            ax,
            vmin=vmin,
            vmax=vmax,
            **kwargs,
        )

        if use_diverging_cmap:
            # for div cmaps usually not the whole range is plotted in the
            # initial plot; therefore we remove the old colorbar and add a new
            # one
            cbar.remove()

            norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=0.0, vmax=vmax)
            cm = ScalarMappable(norm, self.div_cmap)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", "8%", pad="8%")
            cbar = fig.colorbar(cm, cax=cax, label=cbar_label)

            fig.tight_layout()

        # Annotate Phase
        annotation = self._annotate_phase(ax, 0)

        if tripcolor:
            anim = FuncAnimation(
                fig,
                self._animation_func_tripcolor,
                frames=frames,
                interval=interval,
                blit=False,
                fargs=(
                    frames,
                    triplot,
                    annotation,
                    animate_meth,
                    symmetric,
                    use_diverging_cmap,
                ),
            )
        else:
            anim = FuncAnimation(
                fig,
                self._animation_func_tricontourf,
                frames=frames,
                interval=interval,
                blit=False,
                fargs=(
                    frames,
                    ax,
                    triplot,
                    triang,
                    annotation,
                    animate_meth,
                    symmetric,
                    use_diverging_cmap,
                ),
            )

        return anim

    # -------------------------------------------------------------------------
    # Triangulation
    # -------------------------------------------------------------------------

    @staticmethod
    def triangulation(
        X: NDArray,
        Y: NDArray,
        symmetric: bool,
    ) -> tuple[NDArray, NDArray, tri.Triangulation]:
        """Creates the triangulation

        Creates triangulation of points with coordinates (X[i], Y[i])

        :param X: array of x-coordinates
        :param Y: array of y-coordinates
        :param symmetric: if true the plot is symmetric w.r.t. the x-axis
        """
        if symmetric:
            X = np.concatenate([X, X])
            Y = np.concatenate([Y, -Y])

        triangs = tri.Triangulation(X, Y)
        return X, Y, triangs

    def masked_triangulation(
        self,
        X: NDArray,
        Y: NDArray,
        symmetric: bool,
        radius: None | float = None,
    ) -> tri.Triangulation:
        """Creates the masked triangulation

        Creates the masked triangulation of points with coordinates
        (X[i],Y[i]). If ``radius`` is passed points inside radius are masked.

        :param X: array of x-coordinates
        :param Y: array of y-coordinates
        :param symmetric: if true the plot is symmetric w.r.t. the x-axis
        :param radius: radius of the particle
        """
        X, Y, triangs = self.triangulation(X, Y, symmetric=symmetric)
        if not self.is_scatterer:
            X = X[triangs.triangles].mean(axis=1)
            Y = Y[triangs.triangles].mean(axis=1)
            mask = np.hypot(X, Y) < radius
            triangs.set_mask(mask)
        return triangs

    # -------------------------------------------------------------------------
    # Private Methods: Animation
    # -------------------------------------------------------------------------

    def _animation_func_tripcolor(
        self,
        frame: int,
        n_frames: int,
        triplot: TriMesh,
        annotation: Annotation,
        animate_meth: Callable[[float], tuple[NDArray, NDArray, NDArray]],
        symmetric: bool,
        use_diverging_cmap: bool,
    ) -> TriMesh:
        """Animation function for tripcolor plot


        :param frame: frame
        :param n_frames: nbr of frames
        :param triplot: TriMesh object
        :param annotation: phase annotation
        :param animate_meth: function that returns speed for given phase
        :param symmetric: if the symmetric part of plot shall be plotted
        :param use_diverging_cmap: use divergent colormap
        """
        # Compute Phase
        phase = frame / n_frames * 2 * pi
        # Get data
        _, _, speed = animate_meth(phase)
        # Symmetry
        if symmetric:
            speed = np.concatenate([speed, speed])
        # Update plot
        triplot.set_array(speed)

        # Update phase annotation
        self._update_phase_annotation(annotation, phase)

        return triplot

    def _animation_func_tricontourf(
        self,
        frame: int,
        n_frames: int,
        ax: plt.Axes,
        triplot: TriMesh,
        triang: Triangulation,
        annotation: Annotation,
        animate_meth: Callable[[float], tuple[NDArray, NDArray, NDArray]],
        symmetric: bool,
        use_diverging_cmap: bool,
    ) -> TriMesh:
        """Animation function for tricontourf plot

        :param frame: frame
        :param n_frames: nbr of frames
        :param ax: Axes object
        :param triplot: TriMesh object
        :param triang: triangles of triangulation
        :param annotation: phase annotation
        :param animate_meth: function that returns speed for given phase
        :param symmetric: return if the symmetric part of plot shall be plotted
        :param use_diverging_cmap: use divergent colormap
        """
        # Compute Phase
        phase = frame / n_frames * 2 * pi
        # Get data
        _, _, speed = animate_meth(phase)
        # Symmetry
        if symmetric:
            speed = np.concatenate([speed, speed])
        # Get levels
        levels = triplot.levels

        # create symmetric levels if data is positive/negative
        if use_diverging_cmap:
            levels = np.append(levels, -levels[1:])
            levels = np.sort(levels)

        # Remove artists from plot
        collections = triplot.collections
        for coll in collections:
            triplot.collections.remove(coll)

        cmap = self.div_cmap if use_diverging_cmap else self.cmap
        # Plot contours anew
        triplot = ax.tricontourf(triang, speed, cmap=cmap, levels=levels)

        # Update phase annotation
        self._update_phase_annotation(annotation, phase)

        return triplot

    # -------------------------------------------------------------------------
    # Private Methods Axes post processing
    # -------------------------------------------------------------------------

    def _finish_triplot(
        self,
        fig: plt.Figure,
        ax: plt.Axes,
        radius: None | float,
        cbar_mappable: ScalarMappable,
        cbar_label: str,
    ) -> None | Colorbar:
        """Post-processes tricontourf/tripcolor plot

        add ticks, labels, circle, color bar, and color bar label
        :param fig: Figure object
        :param ax: Axes object
        :param radius: particle radius
        :param cbar_mappable: mappable for the color bar
        :param cbar_label: label for the color bar
        """
        self._set_lims(ax)
        self._finish_plot(ax, radius)
        cbar = self._add_colorbar(fig, ax, cbar_mappable, cbar_label)
        return cbar

    @staticmethod
    def _add_colorbar(
        fig: plt.Figure,
        ax: plt.Axes,
        mappable: ScalarMappable,
        label: str,
    ) -> None | Colorbar:
        """Add color bar to figure

        :param fig: figure
        :param mappable: mappable
        """
        if fig is None:
            fig = plt.gcf()  # pragma: no cover

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", "8%", pad="8%")
        cbar = fig.colorbar(mappable, cax=cax, label=label)
        return cbar

    def _set_lims(
        self,
        ax: plt.Axes,
    ) -> None:
        """Set the limits in the plot

        :param ax: Axes object
        """
        ax.autoscale(True, tight=True)
        old_left, old_right = ax.get_xlim()
        if self.is_scatterer:
            left = 1.05 * old_left
            right = 1.05 * old_right
        else:
            left = old_left / sqrt(2)
            right = old_right / sqrt(2)
        ax.set_xlim(left, right)
        ax.set_ylim(left, right)


if __name__ == "__main__":
    pass



from __future__ import annotations

import unittest

import numpy as np

from osaft.core.functions import BesselFunctions as Bf
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory

NDArray = np.ndarray


class TestCoefficientMatrix(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self.cls = SolutionFactory().doinikov_2021_viscoelastic_scattering()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # generic tests
    # -------------------------------------------------------------------------

    def test_viscosity_term(self) -> None:
        self.do_testing(self.cls.viscosity_term, self.viscosity_term)

    def test_D_0(self) -> None:
        self.do_testing(self.cls.D_0, self.D_0)

    def test_a_0(self) -> None:
        self.do_testing(self.cls.a_0, self.a_0)

    def test_a_hat_0(self) -> None:
        self.do_testing(self.cls.a_hat_0, self.a_hat_0)

    def test_matrix_M(self) -> None:
        def element_test_matrix(m: int, i: int, j: int) -> complex:
            return self.cls.matrix_M(m)[i, j]

        def element_matrix(m: int, i: int, j: int) -> complex:
            return self.matrix_M(m)[i, j]

        for ix, iy in np.ndindex(self.cls.matrix_M(0).shape):
            for n in range(0, 3):
                with self.subTest(f"Matrix element {ix}, {iy}, mode {n}"):
                    self.do_testing(
                        element_test_matrix,
                        element_matrix,
                        args_1=(n, ix, iy),
                        args_2=(n, ix, iy),
                        threshold=1e-5,
                    )

    def test_vector_N(self):
        def element_test_vector(m: int, i: int) -> complex:
            return self.vector_N(m)[i]

        def element_vector(m: int, i: int) -> complex:
            return self.cls.vector_n(m)[i]

        for n in range(3):
            for index in range(4):
                with self.subTest(msg=f"coefficient = {n}"):
                    self.do_testing(
                        element_test_vector,
                        element_vector,
                        args_1=(n, index),
                        args_2=(n, index),
                    )

    def test_properties(self) -> None:
        properties = ["x_l", "x_t", "x_f", "x_v"]
        self._test_properties(properties)

    def test_methods_n(self) -> None:
        methods = {}
        methods["a"] = None
        methods["b"] = None
        methods["a_hat"] = None
        methods["b_hat"] = None
        methods["potential_coefficient"] = None
        self._test_methods_n(methods, n_end=6)

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def omega(self):
        return 2 * np.pi * self.f

    @property
    def x_l(self):
        return self.cls.k_l * self.cls.R_0

    @property
    def x_t(self):
        return self.cls.k_t * self.cls.R_0

    @property
    def x_f(self):
        return self.cls.k_f * self.R_0

    @property
    def x_v(self):
        return self.cls.k_v * self.R_0

    @property
    def eta_c(self):
        return self.cls.fluid.eta_c

    @property
    def zeta_c(self):
        return self.cls.fluid.zeta_c

    def viscosity_term(self):
        return (
            1j * self.rho_f * self.c_f**2 / self.omega
            + self.zeta_c
            - 2 * self.eta_c / 3
        )

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    def D_0(self):
        prefactor1 = self.E_s * self.cls.k_f * self.cls.x_l**2
        prefactor1 *= Bf.hankelh1(1, self.x_f) / (1 + self.nu_s)

        term1 = self.nu_s * Bf.besselj(0, self.x_l) / (1 - 2 * self.nu_s)
        term1 -= Bf.d2_besselj(0, self.x_l)

        prefactor2 = 1j * self.omega * self.cls.k_l * self.x_f**2
        prefactor2 *= Bf.besselj(1, self.x_l)

        term2 = Bf.hankelh1(0, self.x_f) * self.viscosity_term()
        term2 -= 2 * self.eta_c * Bf.d2_hankelh1(0, self.x_f)

        return prefactor1 * term1 + prefactor2 * term2

    def a_0(self):

        term1 = Bf.d2_besselj(0, self.x_l)
        term1 -= self.nu_s * Bf.besselj(0, self.x_l) / (1 - 2 * self.nu_s)
        term1 *= self.E_s * self.cls.k_f * self.x_l**2
        term1 *= Bf.besselj(1, self.x_f) / (1 + self.nu_s)

        term2 = 2 * self.eta_c * Bf.d2_besselj(0, self.x_f)
        term2 -= Bf.besselj(0, self.x_f) * self.viscosity_term()
        term2 *= 1j * self.omega * self.cls.k_l * self.x_f**2
        term2 *= Bf.besselj(1, self.x_l)

        return (term1 + term2) * self.cls.A_in(0) / self.D_0()

    def a_hat_0(self):
        return (
            self.cls.A_in(0)
            * self.cls.k_f
            * self.x_f**2
            / self.D_0()
            * (
                Bf.hankelh1(1, self.x_f)
                * (
                    Bf.besselj(0, self.x_f) * self.viscosity_term()
                    - 2 * self.eta_c * Bf.d2_besselj(0, self.x_f)
                )
                + Bf.besselj(1, self.x_f)
                * (
                    2 * self.eta_c * Bf.d2_hankelh1(0, self.x_f)
                    - Bf.hankelh1(0, self.x_f) * self.viscosity_term()
                )
            )
        )

    def matrix_M(self, l: int) -> NDArray:
        M = np.zeros((4, 4), dtype=complex)

        M[0, 0] = self.x_f * Bf.d1_hankelh1(l, self.x_f)
        M[0, 1] = -l * (l + 1) * Bf.hankelh1(l, self.x_v)
        M[0, 2] = 1j * self.omega * self.x_l * Bf.d1_besselj(l, self.x_l)
        M[0, 3] = -1j * self.omega * l * (l + 1) * Bf.besselj(l, self.x_t)
        M[1, 0] = Bf.hankelh1(l, self.x_f)
        M[1, 1] = -Bf.hankelh1(l, self.x_v) - self.x_v * Bf.d1_hankelh1(
            l,
            self.x_v,
        )
        M[1, 2] = 1j * self.omega * Bf.besselj(l, self.x_l)
        M[1, 3] = (
            -1j
            * self.omega
            * (Bf.besselj(l, self.x_t) + self.x_t * Bf.d1_besselj(l, self.x_t))
        )
        M[2, 0] = self.x_f**2 * (
            2 * self.eta_c * Bf.d2_hankelh1(l, self.x_f)
            - self.viscosity_term() * Bf.hankelh1(l, self.x_f)
        )
        M[2, 1] = (
            2
            * l
            * (l + 1)
            * self.eta_c
            * (
                Bf.hankelh1(l, self.x_v)
                - self.x_v * Bf.d1_hankelh1(l, self.x_v)
            )
        )
        M[2, 2] = (
            self.E_s
            * self.x_l**2
            / (1 + self.nu_s)
            * (
                -Bf.d2_besselj(l, self.x_l)
                + self.nu_s * Bf.besselj(l, self.x_l) / (1 - 2 * self.nu_s)
            )
        )
        M[2, 3] = (
            l
            * (l + 1)
            * self.E_s
            / (1 + self.nu_s)
            * (self.x_t * Bf.d1_besselj(l, self.x_t) - Bf.besselj(l, self.x_t))
        )
        M[3, 0] = (
            2
            * self.eta_c
            * (
                self.x_f * Bf.d1_hankelh1(l, self.x_f)
                - Bf.hankelh1(l, self.x_f)
            )
        )
        M[3, 1] = self.eta_c * (
            (2 - l**2 - l) * Bf.hankelh1(l, self.x_v)
            - self.x_v**2 * Bf.d2_hankelh1(l, self.x_v)
        )
        M[3, 2] = (
            self.E_s
            / (1 + self.nu_s)
            * (Bf.besselj(l, self.x_l) - self.x_l * Bf.d1_besselj(l, self.x_l))
        )
        M[3, 3] = (
            self.E_s
            / (2 * (1 + self.nu_s))
            * (
                self.x_t**2 * Bf.d2_besselj(l, self.x_t)
                - (2 - l**2 - l) * Bf.besselj(l, self.x_t)
            )
        )
        return M

    def vector_N(self, l: int) -> NDArray:
        vec_n = np.zeros(4, dtype=complex)

        vec_n[0] = -self.cls.A_in(l) * self.x_f * Bf.d1_besselj(l, self.x_f)
        vec_n[1] = -self.cls.A_in(l) * Bf.besselj(l, self.x_f)
        vec_n[2] = (
            self.cls.A_in(l)
            * self.x_f**2
            * (
                self.viscosity_term() * Bf.besselj(l, self.x_f)
                - 2 * self.eta_c * Bf.d2_besselj(l, self.x_f)
            )
        )
        vec_n[3] = (
            2
            * self.cls.A_in(l)
            * self.eta_c
            * (Bf.besselj(l, self.x_f) - self.x_f * Bf.d1_besselj(l, self.x_f))
        )

        return vec_n

    def det_matrix(self, n: int, column: None | int = None) -> complex:
        matrix = self.matrix_M(n).copy()
        vector = self.vector_N(n)
        if column is not None:
            matrix[:, column] = vector
        return np.linalg.det(matrix)

    def a(self, n: int) -> complex:
        if n == 0:
            return self.a_0()
        else:
            return self.det_matrix(n, 0) / self.det_matrix(n)

    def b(self, n: int) -> complex:
        if n == 0:
            return 0
        else:
            return self.det_matrix(n, 1) / self.det_matrix(n)

    def a_hat(self, n: int) -> complex:
        if n == 0:
            return self.a_hat_0()
        else:
            return self.det_matrix(n, 2) / self.det_matrix(n)

    def b_hat(self, n: int) -> complex:
        if n == 0:
            return 0
        else:
            return self.det_matrix(n, 3) / self.det_matrix(n)

    def potential_coefficient(self, n: int) -> complex:
        return self.a(n)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from collections.abc import Callable

import numpy as np

from osaft import log
from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import clebsch_gordan_coefficient as cg
from osaft.core.functions import conj, full_range, integrate, sqrt
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.doinikov2021viscous.scattering import ScatteringField

NDArray = np.ndarray


class StreamingField(ScatteringField):
    """Streaming field class for Doinikov (viscous fluid-elastic sphere; 2021)

    :param f: frequency [Hz]
    :param R_0: radius [m]
    :param rho_s: density of the particle [kg/m^3]
    :param E_s: Young's modulus of the particle [Pa]
    :param nu_s: Poisson's ratio of the particle [-]
    :param rho_f: density of the fluid [kg/m^3]
    :param c_f: speed of sound in the fluid [m/s]
    :param eta_f: fluid component shear viscosity [Pa s]
    :param zeta_f: fluid component bulk viscosity [Pa s]
    :param p_0: pressure amplitude [Pa]
    :param wave_type: wave type
    :param position: position in the standing wave field [m]
    :param N_max: truncation of the summation index (highest order mode)
    :param inf_factor: see :attr:`.inf_factor`
    :param inf_type: see :attr:`.inf_type`
    :param integration_rel_eps:  relative eps for adaptive integration
    """

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float,
        rho_s: float,
        E_s: float,
        nu_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
        N_max: int = 5,
        inf_factor: float = 60.0,
        inf_type: None | str = None,
        integration_rel_eps: float = 1e-3,
    ) -> None:
        """Initialize class"""

        # __init__ parent class
        ScatteringField.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            E_s=E_s,
            nu_s=nu_s,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
            N_max=N_max,
        )

        # Independent variables
        self._inf_factor = PassiveVariable(inf_factor, "infinity factor")
        self._inf_type = PassiveVariable(
            inf_type,
            "infinity type",
        )

        self._integration_rel_eps = PassiveVariable(
            integration_rel_eps,
            "estimated integration error goal",
        )

        # Dependent variables
        self._C_0 = ActiveVariable(
            self._compute_C_0,
            "integration constants C_0",
        )

        self._inf = ActiveVariable(self._compute_inf, "infinity")

        # Dependencies
        self._C_0.is_computed_by(
            self._inf_factor,
            self._inf_type,
            self._integration_rel_eps,
            self.fluid._rho_f,
            self.fluid._eta_f,
            self.fluid._zeta_f,
            self.fluid._k_f,
        )

        self._inf.is_computed_by(
            self._inf_type,
            self._inf_factor,
            self.sphere._R_0,
            self.fluid._delta,
        )

    # -------------------------------------------------------------------------
    # Getters and Setters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def inf_factor(self) -> float:
        """Infinity factor

        Factor that is multiplied with the length given through
        :attr:`.inf_type` to get the upper limit for integration to the
        far-field.

        :getter: returns the infinity factor
        :rtype: float
        :setter: automatically invokes
            :func:`osaft.core.variable.BaseVariable.notify`
        :type: float
        """

        return self._inf_factor.value

    @inf_factor.setter
    def inf_factor(self, value: float) -> None:
        self._inf_factor.value = value

    @property
    def inf_type(self) -> str:
        """Infinity type

        the upper limit of the integration to the far-field is defined as a
        multiple of either radius, boundary layer thickness, acoustic
        wavelength, viscous wavelength, or a user-defined value.
        The options are
        - `'delta'`
        - `'radius'`
        - `'1'` or 1
        - `None`

        The multiplier is set through
        :attr:`.inf_value`

        :getter: returns the infinity type
        :rtype: str
        :setter: automatically invokes
            :func:`osaft.core.variable.BaseVariable.notify`
        """
        return self._inf_type.value

    @inf_type.setter
    def inf_type(self, value: str) -> None:
        self._inf_type.value = value

    @property
    def integration_rel_eps(self) -> float:
        """Relative integration error goal for adaptive integration

        Relative error goal for the computation of :attr:`.C_integral`

        :getter: returns the relative integration error
        :rtype: float
        :setter: automatically invokes
            :func:`osaft.core.variable.BaseVariable.notify`
        :type: float
        """
        return self._integration_rel_eps.value

    @integration_rel_eps.setter
    def integration_rel_eps(self, value: float) -> None:
        self._integration_rel_eps.value = value

    # -------------------------------------------------------------------------
    # Getters for dependent Variables
    # -------------------------------------------------------------------------

    @property
    def C_0(self) -> NDArray:
        """Container for integrals and errors :math:`C_{ml0}`

        (Eq. E5 - E7)

        Stores the integrals and the estimated errors for the integral
        :math:`C_{ml0}`
        """
        return self._C_0.value["integral"]

    @property
    def C_0_error(self) -> NDArray:
        """Container for integrals and errors :math:`C_{ml0}`

        (Eq. E5 - E7)

        Stores the integrals and the estimated errors for the integral
        :math:`C_{ml0}`
        """
        return self._C_0.value["error"]

    @property
    def inf(self):
        """Limit for the integration to the far-field
        / to infinity) [m]"""
        return self._inf.value

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _integrate_C_ml0(
        self,
        kernel: Callable[[float], float],
    ) -> (float, float):
        return integrate(
            kernel,
            self.R_0,
            self.inf,
            rel_eps=self.integration_rel_eps,
        )

    @staticmethod
    def _compute_C_210(l: int, integral: float) -> float:
        return -1 / (2 * l + 1) * integral

    def _compute_C_210_error(self, l: int, error: float) -> float:
        return self._compute_C_210(l, error)

    def _compute_C_3l0(
        self,
        l: int,
        C_2l0: float,
        C_5l0: float,
        C_6l0: float,
    ) -> float:
        term_1 = self.V_S_r(l, self.R_0, ac=False) / (l + 1)
        term_1 += self.V_S_theta(l, self.R_0, ac=False)
        term_1 *= self.R_0**l

        term_2 = C_2l0 / (l + 1) - C_5l0
        term_2 *= (2 * l + 1) * self.R_0 ** (2 * l - 1)

        term_3 = (2 * l + 3) * self.R_0 ** (2 * l + 1) * C_6l0

        return 0.5 * (term_1 + term_2 - term_3)

    def _compute_C_3l0_error(
        self,
        l: int,
        error_2l0: float,
        error_5l0: float,
        error_6l0: float,
    ) -> float:
        term_1 = error_2l0 / (l + 1) - error_5l0
        term_1 *= (2 * l + 1) * self.R_0 ** (2 * l - 1)
        term_2 = -(2 * l + 3) * self.R_0 ** (2 * l + 1) * error_6l0
        return 0.5 * term_1 - term_2

    def _compute_C_4l0(
        self,
        l: int,
        C_2l0: float,
        C_5l0: float,
        C_6l0: float,
    ) -> float:
        term_1 = (2 - l) * self.V_S_r(l, self.R_0, False) / (l * (l + 1))
        term_1 -= self.V_S_theta(l, self.R_0, ac=False)
        term_1 *= self.R_0 ** (l + 2)

        term_2 = C_2l0 / (l + 1) - C_5l0
        term_2 *= (2 * l - 1) * self.R_0 ** (2 * l + 1)

        term_3 = +(2 * l + 1) * self.R_0 ** (2 * l + 3) * C_6l0

        return 0.5 * (term_1 - term_2 + term_3)

    def _compute_C4l0_error(
        self,
        l: int,
        error_2l0: float,
        error_5l0: float,
        error_6l0: float,
    ) -> float:
        term_1 = (2 * l - 1) * self.R_0 ** (2 * l + 1)
        term_1 *= error_2l0 / (l + 1) - error_5l0

        term_2 = (2 * l + 1) * self.R_0 ** (2 * l + 3) * error_6l0

        return 0.5 * (-term_1 + term_2)

    @staticmethod
    def _compute_C_5l0(l: int, integral: float) -> float:
        return 1 / (2 * (2 * l - 1) * (2 * l + 1)) * integral

    def _compute_C_5l0_error(self, l: int, error: float) -> float:
        return self._compute_C_5l0(l, error)

    @staticmethod
    def _compute_C_6l0(l: int, integral: float) -> float:
        return -1 / (2 * (2 * l + 1) * (2 * l + 3)) * integral

    def _compute_C_6l0_error(self, l: int, error: float) -> float:
        return self._compute_C_6l0(l, error)

    def _compute_C_0(self, L_max: None | int = None) -> dict:
        """Computes the integrals :math:`C_{ml0}` for :math:`l` up to `L_max`

        :param L_max: max index of summation
        :return: dictionary with integrals and errors
        """
        if L_max is None:
            l_range = self.range_1_N_max
        else:
            l_range = full_range(1, L_max)

        integrals = np.zeros((6 + 1, self.N_max + 1))
        errors = np.zeros((6 + 1, self.N_max + 1))

        log.info(
            "Integration to compute C_ml0 is performed. This might take "
            "a while.",
        )

        for l in l_range:

            def kernel_C_2l0(s: float) -> float:
                alpha = self.alpha(l, s, ac=False)
                alpha_incident = self.alpha(l, s, ac=True)
                return s ** (1 - l) * (alpha - alpha_incident)

            def kernel_C_5l0(s: float) -> float:
                E = self.E(l, s, ac=False)
                E_incident = self.E(l, s, ac=True)
                return s ** (3 - l) * (E - E_incident)

            def kernel_C_6l0(s: float) -> float:
                E = self.E(l, s, ac=False)
                E_incident = self.E(l, s, ac=True)
                return s ** (1 - l) * (E - E_incident)

            # C_1l0
            integrals[1, l] = 0
            errors[1, l] = 0

            # C_2l0
            integral_C_2l0, error_C_2l0 = self._integrate_C_ml0(kernel_C_2l0)
            integrals[2, l] = self._compute_C_210(l, integral_C_2l0)
            errors[2, l] = self._compute_C_210_error(l, error_C_2l0)

            # C_5l0
            integral_C_5l0, error_C_5l0 = self._integrate_C_ml0(kernel_C_5l0)
            integrals[5, l] = self._compute_C_5l0(l, integral_C_5l0)
            errors[5, l] = self._compute_C_5l0_error(l, error_C_5l0)

            # C_6l0
            integral_C_6l0, error_C_6l0 = self._integrate_C_ml0(kernel_C_6l0)
            integrals[6, l] = self._compute_C_6l0(l, integral_C_6l0)
            errors[6, l] = self._compute_C_6l0_error(l, error_C_6l0)

            # C_3l0
            integrals[3, l] = self._compute_C_3l0(
                l,
                integrals[2, l],
                integrals[5, l],
                integrals[6, l],
            )
            errors[3, l] = self._compute_C_3l0_error(
                l,
                error_C_2l0,
                error_C_5l0,
                error_C_6l0,
            )

            # C_4l0
            integrals[4, l] = self._compute_C_4l0(
                l,
                integrals[2, l],
                integrals[5, l],
                integrals[6, l],
            )
            errors[4, l] = self._compute_C4l0_error(
                l,
                error_C_2l0,
                error_C_5l0,
                error_C_6l0,
            )

        return {"integral": integrals, "error": errors}

    def _compute_inf(self) -> float:
        if self.inf_type is None:
            if self.R_0 > self.delta:
                return self.inf_factor * self.R_0
            else:
                return self.inf_factor * self.delta
        elif self.inf_type == "delta":
            return self.inf_factor * self.delta
        elif self.inf_type == 1 or self.inf_type == "1":
            return self.inf_factor
        elif self.inf_type == "radius":
            return self.inf_factor * self.R_0
        else:
            raise ValueError(
                f"inf_type needs to be 'boundary layer', '1', "
                f"'radius'"
                f"\n "
                f"got: {self.inf_type}",
            )

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def E(self, l: int, r: float, ac: bool) -> float:
        """:math:`E_l(r)`

        (Eq. C17)

        :param l: index l
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        :return: E(l, r)
        """
        gamma = self.gamma(l, r, ac)
        d_gamma = self.d_gamma(l, r, ac)
        beta = self.beta(l, r, ac)
        return -self.rho_f / (self.eta_f * r) * (gamma + r * d_gamma - beta)

    def F(self, n: int, m: int, r: float, ac: bool) -> complex:
        """:math:`F_{mn}(r)`

        (Eq. B5)

        :param n: index n
        :param m: index m
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        :return: F(n, m, r)
        """
        sc = not ac
        factor = 1j * conj(self.k_f**2) / r
        term_1 = 2 * self.phi(n, r, sc, True).conjugate()
        term_1 += r * self.d_phi(n, r, sc, True).conjugate()
        term_1 *= self.V_r(m, r, sc, True)
        term_2 = r * self.phi(n, r, sc, True).conjugate()
        term_2 *= self.d_V_r(m, r, sc, True)

        return factor * (term_1 + term_2)

    def G(self, n: int, m: int, r: float, ac: bool) -> complex:
        """:math:`G_{mn}(r)`

        (Eq. B6)

        :param n: index n
        :param m: index m
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        :return: G(m, n, r)
        """
        sc = not ac
        factor = 1j * conj(self.k_f**2) / r
        phi = self.phi(n, r, sc, True)
        v_theta = self.V_theta(m, r, sc, True)

        return factor * phi.conjugate() * v_theta

    def V_S_r(self, l: int, r: float, ac: bool) -> float:
        """:math:`V_{Srl}(r)`

        (Eq. F8)

        :param l: index l
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        """

        def first_inner_term(i: int, j: int) -> complex:
            cg_1 = cg(l, 0, j, 0, i, 0)
            if not cg_1:
                return 0
            return cg_1**2 / (2 * i + 1) * conj(self.d_V_r(i, r, sc, True))

        def second_inner_term(i: int, j: int, h: int) -> complex:
            cg_coef = cg(l, 0, j - 2 * i + 1, 0, h, 0)
            if not cg_coef:
                return 0
            factor_1 = (2 * j - 4 * i + 3) * cg_coef**2 / (2 * h + 1)
            factor_2 = self.V_theta(j, r, sc, True) / r
            term_1 = (h + 1) * (h + 2) / (2 * h + 3)
            term_1 *= self.V_r(h + 1, r, sc, True).conjugate()
            if h <= 1:
                return factor_1 * factor_2 * term_1
            term_2 = h * (h - 1) / (2 * h - 1)
            term_2 *= self.V_r(h - 1, r, sc, True).conjugate()
            return factor_1 * factor_2 * (term_1 - term_2)

        sc = not ac

        # First Sum
        first_sum = complex(0)
        for n in self.range_N_max:
            first_inner_sum = complex(0)
            for m in full_range(abs(l - n), l + n):
                first_inner_sum += first_inner_term(m, n)
            first_sum += self.V_r(n, r, sc, True) * first_inner_sum

        # Second Sum
        second_sum = complex(0)
        for n in self.range_1_N_max:
            for m in full_range(1, int((n + 1) / 2)):
                for k in full_range(
                    abs(l - (n - 2 * m + 1)),
                    l + n - 2 * m + 1,
                ):
                    second_sum += second_inner_term(m, n, k)
        return (
            (2 * l + 1)
            / (2 * self.omega)
            * (1j * (first_sum + second_sum)).real
        )

    def V_S_theta(
        self,
        l: int,
        r: float,
        ac: bool,
    ) -> float:
        """:math:`V_{S \\theta l}(r)`

        :param l: index l
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        """

        def first_inner_term(i: int, j: int) -> complex:
            cg_1 = cg(j, 0, l, 0, i, 0)
            cg_2 = cg(j, 0, l, 1, i, 1)
            if not cg_1 or not cg_2:
                return 0
            factor = sqrt(i * (i + 1)) * cg_1 * cg_2 / (2 * i + 1)
            term_1 = self.V_r(j, r, sc, True)
            term_1 *= self.d_V_theta(i, r, sc, True).conjugate()

            term_2 = self.V_r(j, r, sc, True)
            term_2 -= j**2 * self.V_theta(j, r, sc, True)

            term_2 = self.V_theta(i, r, sc, True) * term_2.conjugate() / r

            return factor * (term_1 + term_2)

        # Early return
        if l == 0:
            return 0

        sc = not ac

        # First Sum
        first_sum = complex(0)
        for n in self.range_N_max:
            for m in full_range(abs(n - l), n + l):
                if m >= 1:
                    first_sum += first_inner_term(m, n)
        second_sum = self._V_S_theta_second_sum(l, r, ac)
        return (
            (2 * l + 1)
            / (2 * self.omega * sqrt(l * (l + 1)))
            * (1j * (first_sum + second_sum)).real
        )

    def _V_S_theta_second_sum(self, l: int, r: float, ac: bool) -> complex:
        def second_inner_term(i: int, j: int, h: int) -> complex:
            cg_1 = cg(j - 2 * i, 0, l, 0, h, 0)
            cg_2 = cg(j - 2 * i, 0, l, 1, h, 1)
            if not (h >= 1 and cg_1 and cg_2):
                return 0
            else:
                factor_1 = 2 * j - 4 * i + 1
                factor_2 = sqrt(h * (h + 1)) * cg_1 * cg_2 / (2 * h + 1)
                V_theta = self.V_theta(h, r, sc, True)
                return factor_1 * factor_2 * V_theta

        # Early return
        if l == 0:
            return 0

        sc = not ac

        # Second Sum
        second_sum = complex(0)
        for n in self.range_2_N_max:
            second_inner_sum = complex(0)
            for m in full_range(1, int(n / 2)):
                for k in full_range(abs(n - 2 * m - l), n - 2 * m + l):
                    second_inner_sum += second_inner_term(m, n, k)
            second_sum += (
                conj(self.V_theta(n, r, sc, True)) / r * second_inner_sum
            )
        return second_sum

    def alpha(self, l: int, r: float, ac: bool) -> float:
        """:math:`\\alpha_l(r)`

        (Eq. B9)

        :param l: index l
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        :return: alpha(l, r)
        """

        def first_term(i: int, j: int) -> complex:
            cg_1 = cg(l, 0, j, 0, i, 0)
            if not cg_1:
                return 0
            factor = cg_1**2 / (2 * i + 1)
            F = self.F(j, i, r, ac)
            G = self.G(j, i, r, ac)
            return factor * (F - i * (i + 1) * G)

        def second_term(i: int, j: int, h: int):
            cg_1 = cg(l, 0, j - 2 * i + 1, 0, h, 0)
            if not cg_1:
                return 0
            factor = (2 * j - 4 * i + 3) * cg_1**2 / (2 * h + 1)
            term_1 = (h + 1) * (h + 2) / (2 * h + 3)
            term_1 *= self.G(j, h + 1, r, ac)
            if h <= 1:
                return factor * term_1
            term_2 = h * (h - 1) / (2 * h - 1)
            term_2 *= self.G(j, h - 1, r, ac)
            return factor * (term_1 - term_2)

        # First Sum
        first_sum = complex(0)
        for n in self.range_N_max:
            for m in full_range(abs(l - n), n + l):
                first_sum += first_term(m, n)

        # Second Sum
        second_sum = complex(0)
        for n in self.range_1_N_max:
            for m in full_range(1, int((n + 1) / 2)):
                for k in full_range(
                    abs(l - (n - 2 * m + 1)),
                    l + n - 2 * m + 1,
                ):
                    second_sum += second_term(m, n, k)

        return (2 * l + 1) / (2 * self.omega) * (first_sum + second_sum).real

    def beta(self, l: int, r: float, ac: bool) -> float:
        """:math:`\\beta_l(r)`

        (Eq. C5)

        :param l: index l
        :param r: radial coordinate r
        :param ac: if True only the contribution from the
        :return: beta(l, r)
        """

        def first_term(i: int, j: int) -> float:
            cg_1 = cg(l, 0, j, 0, i, 0)
            if not cg_1:
                return 0
            factor = cg_1**2 / (2 * i + 1) * self.V_r(j, r, sc, True)
            d_V_r = self.d_V_r(i, r, sc, True)
            phi = self.phi(i, r, sc, True)
            return (factor * (d_V_r - self.k_f**2 * phi).conjugate()).real

        def second_term(i: int, j: int, h: int) -> complex:
            cg_1 = cg(l, 0, j - 2 * i + 1, 0, h, 0)
            if not cg_1:
                return 0
            factor = cg_1**2 / (2 * h + 1)
            term_1 = self.V_r(h + 1, r, sc, True)
            term_1 -= self.V_theta(h + 1, r, sc, True)
            term_1 = (h + 1) * (h + 2) / (2 * h + 3) * term_1.conjugate()
            if h <= 1:
                return factor * term_1
            term_2 = self.V_r(h - 1, r, sc, True)
            term_2 -= self.V_theta(h - 1, r, sc, True)
            term_2 = h * (h - 1) / (2 * h - 1) * term_2.conjugate()

            return factor * (term_1 - term_2)

        sc = not ac

        # First Sum
        first_sum = 0
        for n in self.range_N_max:
            first_inner_sum = 0
            for m in full_range(abs(l - n), l + n):
                first_inner_sum += first_term(m, n)
            first_sum += first_inner_sum

        # Second Sum
        second_sum = 0
        for n in self.range_1_N_max:
            for m in full_range(1, int((n + 1) / 2)):
                second_inner_sum = complex(0)
                for k in full_range(
                    abs(l - (n - 2 * m + 1)),
                    l + n - 2 * m + 1,
                ):
                    second_inner_sum += second_term(m, n, k)
                second_inner_sum *= 2 * n - 4 * m + 3
                second_inner_sum *= self.V_theta(n, r, sc, True) / r
                second_sum += second_inner_sum.real
        return (2 * l + 1) / 2 * (first_sum + second_sum)

    def gamma(self, l: int, r: float, ac: bool) -> float:
        """:math:`\\gamma_l(l, r)`

        (Eq. C9)

        :param l: index l
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        :return: gamma(l, r)
        """

        def first_term(i: int, j: int) -> complex:
            cg_1 = cg(j, 0, l, 0, i, 0)
            cg_2 = cg(j, 0, l, 1, i, 1)
            if not cg_1 or not cg_2:
                return 0

            factor = sqrt(i * (i + 1)) * cg_1 * cg_2 / (2 * i + 1)

            term_1 = self.V_r(j, r, sc, True)
            term_1 *= self.d_V_theta(i, r, sc, True).conjugate()

            term_2 = self.V_r(j, r, sc, True) / r
            term_2 -= j**2 * self.V_theta(j, r, sc, True) / r
            term_2 -= self.k_f**2 * self.phi(j, r, sc, True)
            term_2 = self.V_theta(i, r, sc, True) * term_2.conjugate()

            return factor * (term_1 + term_2).real

        if l == 0:
            raise ValueError("gamma for n=0 is not defined")

        sc = not ac

        # First Sum
        first_sum = 0
        for n in self.range_N_max:
            for m in full_range(abs(n - l), n + l):
                if m >= 1:
                    first_sum += first_term(m, n)
        second_sum = self._V_S_theta_second_sum(l, r, ac).real
        return (2 * l + 1) / (2 * sqrt(l * (l + 1))) * (first_sum + second_sum)

    def d_gamma(self, l: int, r: float, ac: bool) -> float:
        """:math:`\\partial_r \\gamma_l(r)`

        (Eq. C9)

        :param l: index l
        :param r: radial coordinate [m]
        :param ac: if True only the incident field contribution
        :return: d_gamma(l, r)
        """

        def first_term(i: int, j: int) -> float:
            cg_1 = cg(j, 0, l, 0, i, 0)
            cg_2 = cg(j, 0, l, 1, i, 1)
            if not cg_1 or not cg_2:
                return 0
            factor = sqrt(i * (i + 1)) * cg_1 * cg_2 / (2 * i + 1)
            term_1 = self.d_V_r(j, r, sc, True)
            term_1 *= self.d_V_theta(i, r, sc, True).conjugate()

            term_2 = self.V_r(j, r, sc, True)
            term_2 *= self.d2_V_theta(i, r, sc, True).conjugate()

            term_3 = self.V_r(j, r, sc, True) / r
            term_3 -= j**2 * self.V_theta(j, r, sc, True) / r
            term_3 -= self.k_f**2 * self.phi(j, r, sc, True)
            term_3 = self.d_V_theta(i, r, sc, True) * term_3.conjugate()

            term_4 = self.d_V_r(j, r, sc, True) / r
            term_4 -= j**2 * self.d_V_theta(j, r, sc, True) / r
            term_4 -= self.V_r(j, r, sc, True) / r**2
            term_4 += j**2 * self.V_theta(j, r, sc, True) / r**2
            term_4 -= self.k_f**2 * self.d_phi(j, r, sc, True)
            term_4 = self.V_theta(i, r, sc, True) * term_4.conjugate()

            return (factor * (term_1 + term_2 + term_3 + term_4)).real

        def second_term(i: int, j: int, h: int) -> float:
            cg_1 = cg(j - 2 * i, 0, l, 0, h, 0)
            cg_2 = cg(j - 2 * i, 0, l, 1, h, 1)
            if not cg_1 or not cg_2:
                return 0
            factor = sqrt(h * (h + 1)) * cg_1 * cg_2 / (2 * h + 1)
            term_1 = self.d_V_theta(j, r, sc, True).conjugate()
            term_1 *= self.V_theta(h, r, sc, True) / r

            term_2 = self.V_theta(j, r, sc, True).conjugate()
            term_2 *= self.d_V_theta(h, r, sc, True) / r

            term_3 = self.V_theta(j, r, sc, True).conjugate()
            term_3 *= self.V_theta(h, r, sc, True) / r**2

            return (factor * (term_1 + term_2 - term_3)).real

        if l == 0:
            raise ValueError("d_gamma for n=0 is not defined")

        sc = not ac

        # First Sum
        first_sum = 0
        for n in self.range_N_max:
            for m in full_range(abs(n - l), n + l):
                if m >= 1:
                    first_sum += first_term(m, n)

        # Second Sum
        second_sum = 0
        for n in self.range_2_N_max:
            for m in full_range(1, int(n / 2)):
                second_inner_sum = 0
                for k in full_range(abs(n - 2 * m - l), n - 2 * m + l):
                    if k >= 1:
                        second_inner_sum += second_term(m, n, k)
                second_sum += (2 * n - 4 * m + 1) * second_inner_sum

        return (2 * l + 1) / (2 * sqrt(l * (l + 1))) * (first_sum + second_sum)


if __name__ == "__main__":
    pass



from __future__ import annotations

from collections.abc import Sequence
from enum import Enum
from numbers import Number

import numpy as np

from osaft.core.functions import pi

NDArray = np.ndarray


class StringFormatter:
    """Consistent Formatting for __str__ and __repr__ methods"""

    @staticmethod
    def get_str_text(
        description: str,
        variable: str,
        value: float | int | str | Enum,
        unit: None | str = None,
        linebreak: bool = True,
    ) -> str:
        """
        Returns a constant formatting for the `__str__` method of the ARF
        classes

        :param description: Physical name for the variable
        :param variable: symbol/name of the variable
        :param value: value of the variable
        :param unit: unit of the variable
        :param linebreak: Appends a linebreak to the end of the text
        """
        out = f"{description:<20s}:\t"
        out += f"{variable:>8s} = "
        if isinstance(value, bool) or value is None:
            out += str(value)
        elif isinstance(value, Enum):
            out += f"{value.name:8s}"
        else:
            out += f"{value:5.2e}"
        if unit is not None and unit != "":
            out += f" [{unit:s}]"
        if linebreak:
            out += "\n"

        return out


class InputHandler:
    """Handles inputs for velocity field methods

    Tests if inputs r, theta, t passed to methods that compute the velocity
    field are valid inputs and converts the inputs to numpy.ndarray.
    """

    @staticmethod
    def _test_array(
        array: NDArray,
        low: None | float = None,
        high: None | float = None,
    ):

        if low is not None and np.any(array < low):
            msg = f"at least one value is smaller than {low}"
            raise ValueError(msg)
        if high is not None and np.any(array > high):
            msg = f"at least one value is greater than {high}"
            raise ValueError(msg)

    @classmethod
    def handle_input(
        cls,
        r: Number | Sequence,
        theta: Number | Sequence,
        t: Number | Sequence,
        R_0: Number | Sequence,
        inside_sphere: bool,
    ):
        """Tests if inputs r, theta, t passed to methods that compute the
        velocity field are valid inputs and converts the
        inputs to numpy.ndarray.

        :param r: radial coordinate
        :param theta: tangential coordinate
        :param t: time
        :param R_0: radius of the sphere
        :param inside_sphere: if `True`/`False` tests if `r` is inside/outside
        """

        # Convert to numpy array if necessary
        arr_r = np.asarray(r)
        arr_theta = np.asarray(theta)
        arr_t = np.asarray(t)

        # Test radius
        if inside_sphere:
            cls._test_array(arr_r, low=0, high=R_0)
        else:
            cls._test_array(arr_r, low=R_0)

        # Test theta
        cls._test_array(arr_theta, 0, pi)

        return arr_r, arr_theta, arr_t


if __name__ == "__main__":
    pass



from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestScattering(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self.cls = SolutionFactory().hasegawa_1969_base()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    @property
    def lambda_rho(self) -> float:
        return self.rho_s / self.rho_f

    @property
    def x_f(self) -> float:
        return self.R_0 * self.cls.k_f

    @property
    def x_s_l(self) -> float:
        return self.R_0 * self.cls.k_s_l

    @property
    def x_s_t(self) -> float:
        return self.R_0 * self.cls.k_s_t

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_properties(self) -> None:
        properties = ["lambda_rho", "x_s_l", "x_s_t", "x_f"]
        self._test_properties(properties)



from __future__ import annotations

from collections.abc import Callable
from functools import partial

import numpy as np
from matplotlib import colors
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.cm import ScalarMappable
from matplotlib.quiver import Quiver

from osaft.core.functions import pi, sqrt
from osaft.plotting.datacontainers.scattering_datacontainer import (
    FluidScatteringData,
)
from osaft.plotting.scattering.base_scattering_plots import (
    BaseScatteringPlots,
)
from osaft.plotting.scattering.tri_plotter import TriangulationPlotter
from osaft.solutions.base_scattering import BaseScattering

NDArray = np.ndarray


class FluidScatteringPlot(BaseScatteringPlots):
    """Class for plotting scattering field of the fluid

    Plots the acoustic field in the fluid around the scatterer using Matplotlib
    tricontourf or tripcolor plotting methods.

    :param sol: solution to be plotted
    :param r_max: radial limit of plot range
    :param resolution: if tuple (radial resolution, tangential resolution)
    :param n_quiver_points: anchor points along z for quiver
    :param cmap: color map
    :param div_cmap: diverging color map
    """

    def __init__(
        self,
        sol: BaseScattering,
        r_max: float,
        resolution: int | tuple[int, int] = 100,
        n_quiver_points: int = 21,
        cmap: str = "winter",
        div_cmap: str = "coolwarm",
    ):
        """Constructor method"""
        self.data = FluidScatteringData(
            sol,
            sqrt(2) * r_max,
            res=resolution,
            n_quiver_points=n_quiver_points,
        )
        self.plotter = TriangulationPlotter(False, cmap, div_cmap)

    # -------------------------------------------------------------------------
    # Attributes
    # -------------------------------------------------------------------------

    @property
    def cmap(self) -> str:
        """Colormap for plotting

        :getter: return the colormap choice
        :setter: sets the colormap choice
        """
        return self.plotter.cmap

    @cmap.setter
    def cmap(self, value: str) -> None:
        self.plotter.cmap = value

    @property
    def div_cmap(self) -> str:
        """Diverging Colormap for plotting

        :getter: return the diverging colormap choice
        :setter: sets the diverging colormap choice
        """
        return self.plotter.div_cmap

    @div_cmap.setter
    def div_cmap(self, value: str) -> None:
        self.plotter.div_cmap = value

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def plot_velocity_potential(
        self,
        inst: bool = True,
        phase: float = 0,
        symmetric: bool = True,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf plot for acoustic velocity potential

        Plots the velocity amplitude of the first-order acoustic velocity field
        of the fluid using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param phase: phase :math:`[0, 2\\pi]`
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """
        data = self.data.get_velocity_potential(
            inst,
            phase,
            mode,
            scattered,
            incident,
        )

        abs_max = np.abs(data[-1]).max()
        if inst and abs_max != 0:
            norm = colors.TwoSlopeNorm(
                vmin=-abs_max,
                vcenter=0.0,
                vmax=abs_max,
            )
            use_div_cmap = True
        else:
            norm = colors.Normalize(vmin=0, vmax=abs_max)
            use_div_cmap = False

        return self._triangulation_plot(
            data=data,
            cbar_label="Acoustic Velocity Potential [m^2/s]",
            tripcolor=tripcolor,
            mode=mode,
            symmetric=symmetric,
            use_diverging_cmap=use_div_cmap,
            ax=ax,
            norm=norm,
            **kwargs,
        )

    def plot_pressure(
        self,
        inst: bool = True,
        phase: float = 0,
        symmetric: bool = True,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf plot for acoustic pressure

        Plots the velocity amplitude of the first-order acoustic velocity field
        of the fluid using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param phase: phase :math:`[0, 2\\pi]`
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """
        data = self.data.get_pressure(
            inst,
            phase,
            mode,
            scattered,
            incident,
        )

        abs_max = np.abs(data[-1]).max()
        if inst and abs_max != 0:
            norm = colors.TwoSlopeNorm(
                vmin=-abs_max,
                vcenter=0.0,
                vmax=abs_max,
            )
            use_div_cmap = True
        else:
            norm = colors.Normalize(vmin=0, vmax=abs_max)
            use_div_cmap = False

        return self._triangulation_plot(
            data=data,
            cbar_label="Acoustic Pressure [Pa]",
            tripcolor=tripcolor,
            mode=mode,
            symmetric=symmetric,
            use_diverging_cmap=use_div_cmap,
            ax=ax,
            norm=norm,
            **kwargs,
        )

    def plot_velocity(
        self,
        inst: bool = True,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        quiver_color: None | str = None,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf plot for acoustic velocity field of the fluid

        Plots the velocity amplitude of the first-order acoustic velocity field
        of the fluid using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param phase: phase ``[0, 2 * pi]``
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """
        data = self.data.get_velocity_magnitude(
            inst,
            phase,
            mode,
            scattered,
            incident,
        )

        fig, ax = self._triangulation_plot(
            data=data,
            cbar_label="Acoustic Velocity [m/s]",
            tripcolor=tripcolor,
            mode=mode,
            symmetric=symmetric,
            use_diverging_cmap=False,
            ax=ax,
            **kwargs,
        )

        if quiver_color:
            self._overlay_quiver(
                ax,
                phase,
                scattered,
                incident,
                mode,
                symmetric,
                quiver_color,
                animation=False,
            )

        return fig, ax

    def animate_pressure(
        self,
        frames: int = 64,
        interval: float = 100.0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> FuncAnimation:
        """Tricontourf animation for acoustic pressure of the fluid

        Animates the pressure of the first-order acoustic velocity
        field of the fluid over one period using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``

        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """
        data_meth = self.data.get_pressure

        return self._triangulation_animation(
            data_meth,
            tripcolor=tripcolor,
            frames=frames,
            interval=interval,
            mode=mode,
            scattered=scattered,
            incident=incident,
            symmetric=symmetric,
            ax=ax,
            cbar_label="Acoustic Pressure [Pa]",
            use_diverging_cmap=True,
            **kwargs,
        )

    def animate_velocity_potential(
        self,
        frames: int = 64,
        interval: float = 100.0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> FuncAnimation:
        """Tricontourf animation for acoustic velocity potential of the fluid

        Animates the velocity potential of the first-order acoustic velocity
        field of the fluid over one period using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``

        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """
        data_meth = self.data.get_velocity_potential

        return self._triangulation_animation(
            data_meth,
            tripcolor=tripcolor,
            frames=frames,
            interval=interval,
            mode=mode,
            scattered=scattered,
            incident=incident,
            symmetric=symmetric,
            ax=ax,
            cbar_label="Acoustic Velocity Potential [m^2/s]",
            use_diverging_cmap=True,
            **kwargs,
        )

    def animate_velocity(
        self,
        frames: int = 64,
        interval: float = 100.0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        quiver_color: None | str = None,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> FuncAnimation:
        """Tricontourf animation for acoustic velocity field of the fluid

        Animates the velocity amplitude of the first-order acoustic velocity
        field of the fluid over one period using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``

        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param quiver_color: if not ``None``, quiver plot
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """

        if quiver_color:
            fig, ax = self.plot_velocity(
                inst=False,
                mode=mode,
                scattered=scattered,
                incident=incident,
                symmetric=symmetric,
            )

            Q = self._overlay_quiver(
                ax,
                0,
                scattered,
                incident,
                mode,
                symmetric,
                animation=True,
                quiver_color=quiver_color,
            )
            annotation = self.plotter._annotate_phase(ax, 0)

            def animate_quiver(frame: int, frames: int, Q: Quiver, annotation):
                phase = frame / frames * 2 * pi

                _, _, u, w, _ = self.data.get_velocity_quiver(
                    phase,
                    mode,
                    scattered,
                    incident,
                    symmetric,
                    animation=True,
                )

                self.plotter._update_phase_annotation(annotation, phase)

                Q.set_UVC(u, w)

                return (Q,)

            anim = FuncAnimation(
                fig,
                animate_quiver,
                frames=frames,
                interval=interval,
                blit=False,
                fargs=(frames, Q, annotation),
            )

            return anim
        else:
            data_meth = self.data.get_velocity_magnitude

            return self._triangulation_animation(
                data_meth,
                tripcolor=tripcolor,
                frames=frames,
                interval=interval,
                mode=mode,
                scattered=scattered,
                incident=incident,
                symmetric=symmetric,
                ax=ax,
                cbar_label="Acoustic Velocity [m/s]",
                use_diverging_cmap=False,
                **kwargs,
            )

    def plot_pressure_evolution(
        self,
        inst: bool = True,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        tripcolor: bool = False,
        layout: tuple[int, int] = (3, 3),
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf for acoustic pressure evolution of the fluid

        Plots the pressure amplitude of the first-order acoustic pressure field
        of the fluid over one period at different phases using
        Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``.

        The first phase value is always :math:`0\\pi` and the last one
        :math:`2\\pi`. The total number of plots and, hence, also the steps
        between the different phase values is the defined by the product of the
        ``layout`` tuple.

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param layout: number of rows and columns for plotting
        :param kwargs: passed through to the parent subplots command
        """

        n_row, n_col, phases = self._evo_compute_layout(layout)

        fig, axes = self._evo_create_subplots(n_row, n_col, **kwargs)

        get_data = partial(
            self.data.get_pressure,
            incident=incident,
            scattered=scattered,
        )

        # Get velocity norm, needed to make colormap of the right range
        X, Z, C_norm = get_data(instantaneous=False, mode=mode, phase=0)

        # Only values inside the plotting range
        C_norm = np.where(
            np.hypot(X, Z) < self.data.r_max,
            C_norm,
            0,
        )

        vmax = 1.01 * np.abs(C_norm).max()
        vmin = -vmax

        # Color bar label
        cbar_label = "Acoustic Pressure [Pa]"

        for i, phase in enumerate(phases):
            row = i // n_col
            col = i % n_col
            ax = axes.flat[i]

            X, Z, C = get_data(instantaneous=True, mode=mode, phase=phase)
            self._evo_plot(
                X=Z,
                Y=X,
                C=C,
                symmetric=symmetric,
                tripcolor=tripcolor,
                cbar_label=cbar_label,
                use_diverging_cmap=True,
                ax=ax,
                vmin=vmin,
                vmax=vmax,
            )
            self._evo_clean_axis(ax, phase, row, col, n_row)

        norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
        self._evo_create_cbar(
            fig,
            axes,
            cbar_label,
            self.plotter.div_cmap,
            norm,
        )

        return fig, axes

    def plot_velocity_potential_evolution(
        self,
        inst: bool = True,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        tripcolor: bool = False,
        layout: tuple[int, int] = (3, 3),
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf for acoustic velocity potential evolution of the fluid

        Plots the velocity potential amplitude of the first-order acoustic
        velocity potential field of the fluid over one period at different
        phases using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``.

        The first phase value is always :math:`0\\pi` and the last one
        :math:`2\\pi`. The total number of plots and, hence, also the steps
        between the different phase values is the defined by the product of the
        ``layout`` tuple.

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param layout: number of rows and columns for plotting
        :param kwargs: passed through to the parent subplots command
        """
        n_row, n_col, phases = self._evo_compute_layout(layout)

        fig, axes = self._evo_create_subplots(n_row, n_col, **kwargs)

        get_data = partial(
            self.data.get_velocity_potential,
            incident=incident,
            scattered=scattered,
        )

        X, Z, C_norm = get_data(instantaneous=False, mode=mode, phase=0)

        # Only values inside the plotting range
        C_norm = np.where(
            np.hypot(X, Z) < self.data.r_max,
            C_norm,
            0,
        )

        vmax = 1.01 * np.abs(C_norm).max()
        vmin = -vmax

        # Color bar label
        cbar_label = "Acoustic Velocity Potential [m^2/s]"

        for i, phase in enumerate(phases):
            row = i // n_col
            col = i % n_col
            ax = axes.flat[i]

            X, Z, C = get_data(instantaneous=True, mode=mode, phase=phase)
            self._evo_plot(
                X=Z,
                Y=X,
                C=C,
                symmetric=symmetric,
                tripcolor=tripcolor,
                cbar_label=cbar_label,
                use_diverging_cmap=True,
                ax=ax,
                vmin=vmin,
                vmax=vmax,
            )
            self._evo_clean_axis(ax, phase, row, col, n_row)

        norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
        self._evo_create_cbar(
            fig,
            axes,
            cbar_label,
            self.plotter.div_cmap,
            norm,
        )

        return fig, axes

    def plot_velocity_evolution(
        self,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        quiver_color: None | str = None,
        tripcolor: bool = False,
        layout: tuple[int, int] = (3, 3),
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf for acoustic velocity field evolution of the fluid

        Plots the velocity amplitude of the first-order acoustic velocity
        field of the fluid over one period at different phases using
        Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot
        .tripcolor.html>`_
        if ``tripcolor = True``.

        The first phase value is always :math:`0\\pi` and the last one
        :math:`2\\pi`. The total number of plots and, hence, also the steps
        between the different phase values is the defined by the product of the
        ``layout`` tuple.

        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param quiver_color: color of the quiver arrows, if None: no arrows
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param layout: number of rows and columns for plotting
        :param kwargs: passed through to the parent subplots command
        """
        n_row, n_col, phases = self._evo_compute_layout(layout)

        fig, axes = self._evo_create_subplots(n_row, n_col, **kwargs)

        get_data = partial(
            self.data.get_velocity_magnitude,
            incident=incident,
            scattered=scattered,
        )

        X, Z, C_norm = get_data(instantaneous=False, mode=mode, phase=0)

        # Only values inside the plotting range
        C_norm = np.where(
            np.hypot(X, Z) < self.data.r_max,
            C_norm,
            0,
        )

        vmin = 0
        vmax = 1.01 * C_norm.max()

        # Color bar label
        cbar_label = "Acoustic Velocity [m/s]"

        for i, phase in enumerate(phases):

            row = i // n_col
            col = i % n_col
            ax = axes.flat[i]

            X, Z, C = get_data(instantaneous=True, mode=mode, phase=phase)
            self._evo_plot(
                X=Z,
                Y=X,
                C=C,
                symmetric=symmetric,
                tripcolor=tripcolor,
                cbar_label=cbar_label,
                use_diverging_cmap=False,
                ax=ax,
                vmin=vmin,
                vmax=vmax,
            )
            if quiver_color:
                self._overlay_quiver(
                    ax,
                    phase,
                    scattered,
                    incident,
                    mode,
                    symmetric,
                    animation=False,
                    quiver_color=quiver_color,
                )
            self._evo_clean_axis(ax, phase, row, col, n_row)

        norm = colors.Normalize(vmin=vmin, vmax=vmax)
        self._evo_create_cbar(fig, axes, cbar_label, self.plotter.cmap, norm)

        return fig, axes

    # -------------------------------------------------------------------------
    # Private Methods for evolution plots
    # -------------------------------------------------------------------------
    def _evo_plot(
        self,
        X: NDArray,
        Y: NDArray,
        C: NDArray,
        symmetric: bool,
        tripcolor: bool,
        cbar_label: str,
        use_diverging_cmap: bool,
        ax: plt.Axes,
        vmin: float,
        vmax: float,
    ) -> None:
        # Plot
        _, _, _, cbar, _ = self.plotter.plot(
            X=X,
            Y=Y,
            C=C,
            radius=self.data.sol.R_0,
            symmetric=symmetric,
            tripcolor=tripcolor,
            cbar_label=cbar_label,
            use_diverging_cmap=use_diverging_cmap,
            ax=ax,
            vmin=vmin,
            vmax=vmax,
        )

        cbar.remove()

    @staticmethod
    def _evo_create_cbar(
        fig: plt.Figure,
        axes: plt.Axes,
        cbar_label: str,
        cmap: str,
        norm: colors.Normalize,
    ) -> None:
        fig.tight_layout()
        cm = ScalarMappable(norm, cmap)
        cbar = fig.colorbar(cm, ax=axes.ravel().tolist())
        cbar.ax.set_ylabel(cbar_label)

    @staticmethod
    def _evo_create_subplots(
        n_row: int,
        n_col: int,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        fig, axes = plt.subplots(
            n_row,
            n_col,
            sharex=True,
            sharey=True,
            **kwargs,
        )
        return fig, axes

    @staticmethod
    def _evo_compute_layout(
        layout: tuple[int, int],
    ) -> tuple[int, int, NDArray]:
        n_row, n_col = layout
        n = n_col * n_row

        phases = np.linspace(0, 2 * np.pi, num=n)
        return n_row, n_col, phases

    def _evo_clean_axis(
        self,
        ax: plt.Axes,
        phase: float,
        row: int,
        col: int,
        n_row: int,
    ) -> None:
        ticks = self.data.sol.R_0 * np.asarray([-1, 1])
        labels = [-1, 1]

        ax.set_title(f"{phase / np.pi:.2f}" + r"$\pi$")

        if row != (n_row - 1):
            ax.set_xlabel("")
        if col > 0:
            ax.set_ylabel("")

        ax.set_xticks(ticks, labels=labels)
        ax.set_yticks(ticks, labels=labels)

        # set aspect ratio to 1:1
        ax.set_aspect(1)

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _triangulation_plot(
        self,
        data: tuple[NDArray, NDArray, NDArray],
        cbar_label: str,
        tripcolor: bool,
        symmetric: bool,
        use_diverging_cmap: bool,
        ax: None | plt.Axes,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Helper function for tripcolor/tricontourf plot

        :param data: data to be plotted (X, Z, plot_data)
        :param cbar_label: label for the colorbar
        :param tripcolor: if ``True`` tripcolor, else tricontourf plot
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to plotting method
        """
        kwargs.pop("mode", None)

        X, Z, C = data

        # Plot
        fig, ax, _, _, _ = self.plotter.plot(
            X=Z,
            Y=X,
            C=C,
            radius=self.data.sol.R_0,
            symmetric=symmetric,
            tripcolor=tripcolor,
            cbar_label=cbar_label,
            ax=ax,
            use_diverging_cmap=use_diverging_cmap,
            **kwargs,
        )

        return fig, ax

    def _triangulation_animation(
        self,
        data_meth: Callable[
            [bool, float, int, bool, bool],
            tuple[NDArray, NDArray, NDArray],
        ],
        tripcolor: bool,
        frames: int,
        interval: float,
        mode: None | int,
        scattered: bool,
        incident: bool,
        symmetric: bool,
        cbar_label: str,
        use_diverging_cmap: bool,
        ax: None | plt.Axes,
        **kwargs,
    ) -> FuncAnimation:
        """Helper function for tripcolor/tricontourf animation

        :param data_meth: method for getting the plotting data
        :param tripcolor: if ``True`` tripcolor, else tricontourf plot
        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param scattered: if ``True`` scattering field is plotted
        :param incident: if ``True`` incident field is plotted
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param cbar_label: label of colorbar
        :param use_diverging_cmap: use :attr:`div_cmap` as colormap
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        :param kwargs: passed through to tricontourf()
        """
        # Data function for animation
        animate_meth = lambda phase: data_meth(
            phase=phase,
            instantaneous=True,
            mode=mode,
            scattered=scattered,
            incident=incident,
        )

        # Get velocity norm, needed to make colormap of the right range
        X, Y, C_norm = data_meth(
            instantaneous=False,
            mode=mode,
            scattered=scattered,
            incident=incident,
        )
        # Only values inside the plotting range
        C_norm = np.where(
            np.hypot(X, Y) < self.data.r_max,
            C_norm,
            0,
        )

        return self.plotter.animate(
            frames=frames,
            interval=interval,
            tripcolor=tripcolor,
            symmetric=symmetric,
            cbar_label=cbar_label,
            use_diverging_cmap=use_diverging_cmap,
            animate_meth=animate_meth,
            C_norm=C_norm,
            radius=self.data.sol.R_0,
            ax=ax,
            **kwargs,
        )


if __name__ == "__main__":
    pass



from __future__ import annotations

from osaft import log
from osaft.core.backgroundfields import BackgroundField, WaveType
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.fluids import InviscidFluid
from osaft.core.frequency import Frequency
from osaft.core.functions import pi, sin
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable
from osaft.core.warnings import EPS, raise_assumption_warning
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.base_solution import BaseSolution


class ARF(BaseARF, BaseSphereFrequencyComposite, BaseSolution):
    """ARF class for Gor'kov (1962)

    The standing wave solution is based on equation (13) of the paper and the
    traveling wave solution on equation (10).

    .. note::
       This model is based on the following assumptions:

       - :math:`\\lambda\\gg R`
       - Inviscid fluid
       - Compressible (fluid-like) sphere

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param c_s: Speed of sound of the fluid-like sphere [m/s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of incident wave (traveling/standing)
    :param position: Position in the standing wave field [rad]
    """

    # Supported wave_type for the class
    supported_wavetypes = [WaveType.STANDING, WaveType.TRAVELLING]

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        c_s: float,
        rho_f: float,
        c_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
    ) -> None:
        """Constructor method"""

        # init of parent class
        BaseSphereFrequencyComposite.__init__(self, f, R_0)
        BaseSolution.__init__(self, "Gorkov1962")

        # Initialize Components
        self.scatterer = InviscidFluid(self.frequency, rho_s, c_s)
        self.fluid = InviscidFluid(self.frequency, rho_f, c_f)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        # Dependent variables
        self._f_1 = ActiveVariable(
            self._compute_f_1,
            "monopole scattering coefficient f_1",
        )
        self._f_2 = ActiveVariable(
            self._compute_f_2,
            "dipole scattering coefficient f_2",
        )
        self._Phi = ActiveVariable(
            self._compute_Phi,
            "acoustic contrast factor Phi",
        )

        # Dependencies
        self._f_1.is_computed_by(self.fluid._kappa_f, self.scatterer._kappa_f)
        self._f_2.is_computed_by(self.fluid._rho_f, self.scatterer._rho_f)
        self._Phi.is_computed_by(self._f_1, self._f_2)

        log.info(str(self))
        log.debug(repr(self))

    def __repr__(self):
        return (
            f"GorkovARF({self.f}, {self.R_0}, {self.rho_s}, "
            f"{self.c_s}, {self.rho_f}, {self.c_f}, "
            f"{self.wave_type}, {self.p_0}, {self.position})"
        )

    def __str__(self):
        out = "Gorkov solution with following properties: \n"
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "rad",
        )
        out += SF.get_str_text(
            "Wave Type",
            "",
            self.wave_type,
            None,
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_0",
            self.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_f",
            self.kappa_f,
            "1/Pa",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Speed of Sound",
            "c_s",
            self.scatterer.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_s",
            self.kappa_s,
            "1/Pa",
            False,
        )
        return out

    def compute_arf(self) -> float:
        """Computes the ARF and returns the force in Newton [N].

        Computes ARF according to Eq. (10) or (11)
        Checks before computation of assumption of theory small particle
        radius to pressure field wavelength is valid.

        :raises WrongWaveTypeError: if wrong :attr:`wave_type`
        :raises AssumptionWarning: if the used parameters might not be
            valid for the chosen limiting case
        """
        # Checking wave_type
        self.check_wave_type()

        # Checking particle size section E. in Bruus' paper
        test_value = self.R_0 / self.field.lambda_f
        raise_assumption_warning(test_value < EPS)

        # Case
        if self.wave_type == WaveType.STANDING:
            return self._standing_wave_solution()
        else:
            return self._traveling_wave_solution()

    def _standing_wave_solution(self) -> float:
        out = 4 * pi * self.E_ac * self.R_0**3 * self.k_f
        out *= self.Phi * sin(2 * self.position)
        return out

    def _traveling_wave_solution(self) -> float:
        out = 4 * pi / (9 * self.c_f) * self.I_ac * self.k_f**4
        out *= self.R_0**6
        out *= self.f_1**2 + self.f_1 * self.f_2 + 3 / 4 * self.f_2**2
        return out

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def f_1(self) -> float:
        """Monopole scattering coefficient :math:`f_{1}` [-]"""
        return self._f_1.value

    @property
    def f_2(self) -> float:
        """Dipole scattering coefficient :math:`f_{2}` [-]"""
        return self._f_2.value

    @property
    def Phi(self) -> float:
        """Acoustic contrast factor :math:`\\Phi`  [-]"""
        return self._Phi.value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.p_0`
        """
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> WaveType:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: WaveType) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def E_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.E_ac`
        """
        return self.field.E_ac

    @property
    def I_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.I_ac`
        """
        return self.field.I_ac

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`"""
        return self.scatterer.rho_f

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_f = value

    @property
    def c_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`"""
        return self.scatterer.c_f

    @c_s.setter
    def c_s(self, value: float) -> None:
        self.scatterer.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.kappa_f`"""
        return self.scatterer.kappa_f

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`"""
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`"""
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.kappa_f`"""
        return self.fluid.kappa_f

    @property
    def k_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.k_f`"""
        return self.field.k_f

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_f_1(self) -> float:
        return 1 - self.kappa_s / self.kappa_f

    def _compute_f_2(self) -> float:
        rho_tilde = self.rho_s / self.rho_f
        return 2 * (rho_tilde - 1) / (2 * rho_tilde + 1)

    def _compute_Phi(self):
        return self.f_1 / 3 + self.f_2 / 2


if __name__ == "__main__":
    pass



import inspect
import unittest

import numpy as np
from matplotlib import pyplot as plt

from osaft import ARFPlot
from osaft.solutions.base_arf import BaseARF
from osaft.tests.plotting.basetest_plotting import BaseTestPlotting
from osaft.tests.solution_factory import SolutionFactory


def create_solutions() -> list[BaseARF]:

    sols = [
        SolutionFactory().king_1934_arf(),
        SolutionFactory().yosioka_1955_arf(),
        SolutionFactory().gorkov_1962_arf(),
    ]

    return sols


class TestARFPlot(BaseTestPlotting):
    def setUp(self) -> None:
        super().setUp()

        self.plotter = ARFPlot()

    def test_name_error(self) -> None:
        solutions = create_solutions()
        for sol in solutions:
            self.plotter.add_solutions(sol)
            self.assertRaises(ValueError, self.plotter.add_solutions, sol)

    def test_error_add_zero_solutions(self) -> None:
        self.assertRaises(TypeError, self.plotter.add_solutions)

    def test_add_multiple_solutions(self):
        solutions = create_solutions()
        self.plotter.add_solutions(*solutions)

        # Test if solutions are inside
        for sol in solutions:
            self.assertTrue(sol.name in self.plotter._solutions.keys())

    def test_modulo_in_linestyles(self) -> None:
        solutions = create_solutions()
        for i in np.arange(6):
            for sol in solutions:
                tmp = sol
                tmp.name = f"{tmp.name}_{i}"
                self.plotter.add_solutions(tmp)

        self.plotter.set_abscissa(np.linspace(1e-6, 1e-5), "R_0")

        fig, ax = self.plotter.plot_solutions()

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_add_and_remove_solutions(self) -> None:
        solutions = create_solutions()
        for sol in solutions:
            self.plotter.add_solutions(sol)
        self.plotter.remove_solution(solutions[1])

    def test_repeated_ARF_over_R(self) -> None:
        solutions = create_solutions()

        self.plotter.set_abscissa(np.linspace(1e-6, 1e-5), "R_0")

        for sol in solutions:
            self.plotter.add_solutions(sol)

        fig, ax = self.plotter.plot_solutions(ls="--")
        fig, ax = self.plotter.plot_solutions(linestyle="--")

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_ARF_over_R(self) -> None:
        solutions = create_solutions()

        self.plotter.set_abscissa(np.linspace(1e-6, 1e-5, num=100), "R_0")

        for sol in solutions:
            self.plotter.add_solutions(sol)

        fig, ax = self.plotter.plot_solutions()

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_normARF_over_f(self) -> None:
        solutions = create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), "f")

        for sol in solutions:
            self.plotter.add_solutions(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization=solutions[1].name,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_normARF_over_f_semilogx(self) -> None:
        solutions = create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), "f")

        for sol in solutions:
            self.plotter.add_solutions(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization=solutions[1].name,
            plot_method=plt.semilogx,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_normARF_over_f_loglog(self) -> None:
        solutions = create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), "f")

        for sol in solutions:
            self.plotter.add_solutions(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization=solutions[1].name,
            plot_method=plt.loglog,
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)

    def test_max_norm_over_f(self):
        solutions = create_solutions()

        self.plotter.set_abscissa(np.linspace(1e4, 5e6, num=100), "f")

        for sol in solutions:
            self.plotter.add_solutions(sol)

        fig, ax = self.plotter.plot_solutions(
            normalization="max",
        )

        name = inspect.stack()[0][3]  # method name
        self.save_fig(fig, name)


class DummyClass(BaseARF):
    """Dummy class to test methods"""

    name = "Dummy"
    R_0 = 1

    def compute_arf(self):
        return 2 * self.R_0


class TestARFNormalization(unittest.TestCase):
    def setUp(self) -> None:

        self.cls = DummyClass()
        self.R_values = np.arange(1, 10)

        self.arf_plot = ARFPlot("R_0", self.R_values)
        self.arf_plot.add_solutions(self.cls)

    def test_display_values(self):

        x_values = np.linspace(0, 1, len(self.R_values))
        _, ax = self.arf_plot.plot_solutions(display_values=x_values)
        abscissa = ax.lines[0].get_xdata()
        np.testing.assert_array_equal(x_values, abscissa)

    def test_max_norm(self) -> None:

        _, ax = self.arf_plot.plot_solutions(normalization="max")
        arf = ax.lines[0].get_ydata()
        self.assertAlmostEqual(max(arf), 1)
        self.assertTrue(np.all(arf <= 1))

    def test_other_class_norm(self):
        norm = "Dummy"
        _, ax = self.arf_plot.plot_solutions(normalization=norm)
        arf = ax.lines[0].get_ydata()
        self.assertAlmostEqual(max(arf), 1)

    def test_float_norm(self) -> None:

        norm = 5
        _, ax = self.arf_plot.plot_solutions(normalization=norm)
        arf = ax.lines[0].get_ydata()
        np.testing.assert_array_equal(2 * self.R_values / 5, arf)

    def test_array_norm(self) -> None:

        norm = np.arange(1, 10)
        _, ax = self.arf_plot.plot_solutions(normalization=norm)
        arf = ax.lines[0].get_ydata()
        np.testing.assert_array_equal(2 * np.ones(len(norm)), arf)

    def test_callable_norm(self) -> None:
        def norm(R_0):
            return 2 * R_0

        _, ax = self.arf_plot.plot_solutions(normalization=norm)
        arf = ax.lines[0].get_ydata()
        np.testing.assert_array_equal(np.ones(len(self.R_values)), arf)

    def test_wrong_norm(self):
        self.assertRaises(
            ValueError,
            self.arf_plot.plot_solutions,
            normalization="blabla",
        )


class TestMulticore(unittest.TestCase):
    def test_multicore(self):

        king, _, _ = create_solutions()
        king_copy = king.copy()
        king_copy.name = "King Copy"

        plot = ARFPlot("R_0", np.linspace(1e-6, 10e6))
        plot.add_solutions(king)
        plot.add_solutions(king_copy, multicore=True)

        fig, ax = plot.plot_solutions()

        arf_king = ax.lines[0].get_ydata()
        arf_king_copy = ax.lines[0].get_ydata()

        np.testing.assert_array_equal(arf_king, arf_king_copy)


if __name__ == "__main__":
    unittest.main()
